# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Tsm(KaitaiStruct):
    """InfluxDB is a scalable database optimized for storage of time
    series, real-time application metrics, operations monitoring events,
    etc, written in Go.
    
    Data is stored in .tsm files, which are kept pretty simple
    conceptually. Each .tsm file contains a header and footer, which
    stores offset to an index. Index is used to find a data block for a
    requested time boundary.
    """
    SEQ_FIELDS = ["header"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = Tsm.Header(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()

    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x16\xD1\x16\xD1":
                raise kaitaistruct.ValidationNotEqualError(b"\x16\xD1\x16\xD1", self.magic, self._io, u"/types/header/seq/0")
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u1()
            self._debug['version']['end'] = self._io.pos()


    class Index(KaitaiStruct):
        SEQ_FIELDS = ["offset"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['offset']['start'] = self._io.pos()
            self.offset = self._io.read_u8be()
            self._debug['offset']['end'] = self._io.pos()

        class IndexHeader(KaitaiStruct):
            SEQ_FIELDS = ["key_len", "key", "type", "entry_count", "index_entries"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['key_len']['start'] = self._io.pos()
                self.key_len = self._io.read_u2be()
                self._debug['key_len']['end'] = self._io.pos()
                self._debug['key']['start'] = self._io.pos()
                self.key = (self._io.read_bytes(self.key_len)).decode(u"UTF-8")
                self._debug['key']['end'] = self._io.pos()
                self._debug['type']['start'] = self._io.pos()
                self.type = self._io.read_u1()
                self._debug['type']['end'] = self._io.pos()
                self._debug['entry_count']['start'] = self._io.pos()
                self.entry_count = self._io.read_u2be()
                self._debug['entry_count']['end'] = self._io.pos()
                self._debug['index_entries']['start'] = self._io.pos()
                self.index_entries = [None] * (self.entry_count)
                for i in range(self.entry_count):
                    if not 'arr' in self._debug['index_entries']:
                        self._debug['index_entries']['arr'] = []
                    self._debug['index_entries']['arr'].append({'start': self._io.pos()})
                    _t_index_entries = Tsm.Index.IndexHeader.IndexEntry(self._io, self, self._root)
                    _t_index_entries._read()
                    self.index_entries[i] = _t_index_entries
                    self._debug['index_entries']['arr'][i]['end'] = self._io.pos()

                self._debug['index_entries']['end'] = self._io.pos()

            class IndexEntry(KaitaiStruct):
                SEQ_FIELDS = ["min_time", "max_time", "block_offset", "block_size"]
                def __init__(self, _io, _parent=None, _root=None):
                    self._io = _io
                    self._parent = _parent
                    self._root = _root if _root else self
                    self._debug = collections.defaultdict(dict)

                def _read(self):
                    self._debug['min_time']['start'] = self._io.pos()
                    self.min_time = self._io.read_u8be()
                    self._debug['min_time']['end'] = self._io.pos()
                    self._debug['max_time']['start'] = self._io.pos()
                    self.max_time = self._io.read_u8be()
                    self._debug['max_time']['end'] = self._io.pos()
                    self._debug['block_offset']['start'] = self._io.pos()
                    self.block_offset = self._io.read_u8be()
                    self._debug['block_offset']['end'] = self._io.pos()
                    self._debug['block_size']['start'] = self._io.pos()
                    self.block_size = self._io.read_u4be()
                    self._debug['block_size']['end'] = self._io.pos()

                class BlockEntry(KaitaiStruct):
                    SEQ_FIELDS = ["crc32", "data"]
                    def __init__(self, _io, _parent=None, _root=None):
                        self._io = _io
                        self._parent = _parent
                        self._root = _root if _root else self
                        self._debug = collections.defaultdict(dict)

                    def _read(self):
                        self._debug['crc32']['start'] = self._io.pos()
                        self.crc32 = self._io.read_u4be()
                        self._debug['crc32']['end'] = self._io.pos()
                        self._debug['data']['start'] = self._io.pos()
                        self.data = self._io.read_bytes((self._parent.block_size - 4))
                        self._debug['data']['end'] = self._io.pos()


                @property
                def block(self):
                    if hasattr(self, '_m_block'):
                        return self._m_block if hasattr(self, '_m_block') else None

                    io = self._root._io
                    _pos = io.pos()
                    io.seek(self.block_offset)
                    self._debug['_m_block']['start'] = io.pos()
                    self._m_block = Tsm.Index.IndexHeader.IndexEntry.BlockEntry(io, self, self._root)
                    self._m_block._read()
                    self._debug['_m_block']['end'] = io.pos()
                    io.seek(_pos)
                    return self._m_block if hasattr(self, '_m_block') else None



        @property
        def entries(self):
            if hasattr(self, '_m_entries'):
                return self._m_entries if hasattr(self, '_m_entries') else None

            _pos = self._io.pos()
            self._io.seek(self.offset)
            self._debug['_m_entries']['start'] = self._io.pos()
            self._m_entries = []
            i = 0
            while True:
                if not 'arr' in self._debug['_m_entries']:
                    self._debug['_m_entries']['arr'] = []
                self._debug['_m_entries']['arr'].append({'start': self._io.pos()})
                _t__m_entries = Tsm.Index.IndexHeader(self._io, self, self._root)
                _t__m_entries._read()
                _ = _t__m_entries
                self._m_entries.append(_)
                self._debug['_m_entries']['arr'][len(self._m_entries) - 1]['end'] = self._io.pos()
                if self._io.pos() == (self._io.size() - 8):
                    break
                i += 1
            self._debug['_m_entries']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_entries if hasattr(self, '_m_entries') else None


    @property
    def index(self):
        if hasattr(self, '_m_index'):
            return self._m_index if hasattr(self, '_m_index') else None

        _pos = self._io.pos()
        self._io.seek((self._io.size() - 8))
        self._debug['_m_index']['start'] = self._io.pos()
        self._m_index = Tsm.Index(self._io, self, self._root)
        self._m_index._read()
        self._debug['_m_index']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_index if hasattr(self, '_m_index') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class GlibcUtmp(KaitaiStruct):

    class EntryType(Enum):
        empty = 0
        run_lvl = 1
        boot_time = 2
        new_time = 3
        old_time = 4
        init_process = 5
        login_process = 6
        user_process = 7
        dead_process = 8
        accounting = 9
    SEQ_FIELDS = ["records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['records']['start'] = self._io.pos()
        self._raw_records = []
        self.records = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            self._raw_records.append(self._io.read_bytes(384))
            _io__raw_records = KaitaiStream(BytesIO(self._raw_records[-1]))
            _t_records = GlibcUtmp.Record(_io__raw_records, self, self._root)
            _t_records._read()
            self.records.append(_t_records)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['records']['end'] = self._io.pos()

    class Record(KaitaiStruct):
        SEQ_FIELDS = ["ut_type", "pid", "line", "id", "user", "host", "exit", "session", "tv", "addr_v6", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ut_type']['start'] = self._io.pos()
            self.ut_type = KaitaiStream.resolve_enum(GlibcUtmp.EntryType, self._io.read_s4le())
            self._debug['ut_type']['end'] = self._io.pos()
            self._debug['pid']['start'] = self._io.pos()
            self.pid = self._io.read_u4le()
            self._debug['pid']['end'] = self._io.pos()
            self._debug['line']['start'] = self._io.pos()
            self.line = (self._io.read_bytes(32)).decode(u"UTF-8")
            self._debug['line']['end'] = self._io.pos()
            self._debug['id']['start'] = self._io.pos()
            self.id = (self._io.read_bytes(4)).decode(u"UTF-8")
            self._debug['id']['end'] = self._io.pos()
            self._debug['user']['start'] = self._io.pos()
            self.user = (self._io.read_bytes(32)).decode(u"UTF-8")
            self._debug['user']['end'] = self._io.pos()
            self._debug['host']['start'] = self._io.pos()
            self.host = (self._io.read_bytes(256)).decode(u"UTF-8")
            self._debug['host']['end'] = self._io.pos()
            self._debug['exit']['start'] = self._io.pos()
            self.exit = self._io.read_u4le()
            self._debug['exit']['end'] = self._io.pos()
            self._debug['session']['start'] = self._io.pos()
            self.session = self._io.read_s4le()
            self._debug['session']['end'] = self._io.pos()
            self._debug['tv']['start'] = self._io.pos()
            self.tv = GlibcUtmp.Timeval(self._io, self, self._root)
            self.tv._read()
            self._debug['tv']['end'] = self._io.pos()
            self._debug['addr_v6']['start'] = self._io.pos()
            self.addr_v6 = self._io.read_bytes(16)
            self._debug['addr_v6']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(20)
            self._debug['reserved']['end'] = self._io.pos()


    class Timeval(KaitaiStruct):
        SEQ_FIELDS = ["sec", "usec"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['sec']['start'] = self._io.pos()
            self.sec = self._io.read_s4le()
            self._debug['sec']['end'] = self._io.pos()
            self._debug['usec']['start'] = self._io.pos()
            self.usec = self._io.read_s4le()
            self._debug['usec']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class IcmpPacket(KaitaiStruct):

    class IcmpTypeEnum(Enum):
        echo_reply = 0
        destination_unreachable = 3
        source_quench = 4
        redirect = 5
        echo = 8
        time_exceeded = 11
    SEQ_FIELDS = ["icmp_type", "destination_unreachable", "time_exceeded", "echo"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['icmp_type']['start'] = self._io.pos()
        self.icmp_type = KaitaiStream.resolve_enum(IcmpPacket.IcmpTypeEnum, self._io.read_u1())
        self._debug['icmp_type']['end'] = self._io.pos()
        if self.icmp_type == IcmpPacket.IcmpTypeEnum.destination_unreachable:
            self._debug['destination_unreachable']['start'] = self._io.pos()
            self.destination_unreachable = IcmpPacket.DestinationUnreachableMsg(self._io, self, self._root)
            self.destination_unreachable._read()
            self._debug['destination_unreachable']['end'] = self._io.pos()

        if self.icmp_type == IcmpPacket.IcmpTypeEnum.time_exceeded:
            self._debug['time_exceeded']['start'] = self._io.pos()
            self.time_exceeded = IcmpPacket.TimeExceededMsg(self._io, self, self._root)
            self.time_exceeded._read()
            self._debug['time_exceeded']['end'] = self._io.pos()

        if  ((self.icmp_type == IcmpPacket.IcmpTypeEnum.echo) or (self.icmp_type == IcmpPacket.IcmpTypeEnum.echo_reply)) :
            self._debug['echo']['start'] = self._io.pos()
            self.echo = IcmpPacket.EchoMsg(self._io, self, self._root)
            self.echo._read()
            self._debug['echo']['end'] = self._io.pos()


    class DestinationUnreachableMsg(KaitaiStruct):

        class DestinationUnreachableCode(Enum):
            net_unreachable = 0
            host_unreachable = 1
            protocol_unreachable = 2
            port_unreachable = 3
            fragmentation_needed_and_df_set = 4
            source_route_failed = 5
            dst_net_unkown = 6
            sdt_host_unkown = 7
            src_isolated = 8
            net_prohibited_by_admin = 9
            host_prohibited_by_admin = 10
            net_unreachable_for_tos = 11
            host_unreachable_for_tos = 12
            communication_prohibited_by_admin = 13
            host_precedence_violation = 14
            precedence_cuttoff_in_effect = 15
        SEQ_FIELDS = ["code", "checksum"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = KaitaiStream.resolve_enum(IcmpPacket.DestinationUnreachableMsg.DestinationUnreachableCode, self._io.read_u1())
            self._debug['code']['end'] = self._io.pos()
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_u2be()
            self._debug['checksum']['end'] = self._io.pos()


    class TimeExceededMsg(KaitaiStruct):

        class TimeExceededCode(Enum):
            time_to_live_exceeded_in_transit = 0
            fragment_reassembly_time_exceeded = 1
        SEQ_FIELDS = ["code", "checksum"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = KaitaiStream.resolve_enum(IcmpPacket.TimeExceededMsg.TimeExceededCode, self._io.read_u1())
            self._debug['code']['end'] = self._io.pos()
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_u2be()
            self._debug['checksum']['end'] = self._io.pos()


    class EchoMsg(KaitaiStruct):
        SEQ_FIELDS = ["code", "checksum", "identifier", "seq_num", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['code']['start'] = self._io.pos()
            self.code = self._io.read_bytes(1)
            self._debug['code']['end'] = self._io.pos()
            if not self.code == b"\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x00", self.code, self._io, u"/types/echo_msg/seq/0")
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_u2be()
            self._debug['checksum']['end'] = self._io.pos()
            self._debug['identifier']['start'] = self._io.pos()
            self.identifier = self._io.read_u2be()
            self._debug['identifier']['end'] = self._io.pos()
            self._debug['seq_num']['start'] = self._io.pos()
            self.seq_num = self._io.read_u2be()
            self._debug['seq_num']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes_full()
            self._debug['data']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class GltfBinary(KaitaiStruct):
    """glTF is a format for distribution of 3D models optimized for being used in software
    
    .. seealso::
       Source - https://github.com/KhronosGroup/glTF/tree/2354846/specification/2.0#binary-gltf-layout
    """

    class ChunkType(Enum):
        bin = 5130562
        json = 1313821514
    SEQ_FIELDS = ["header", "chunks"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = GltfBinary.Header(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['chunks']['start'] = self._io.pos()
        self.chunks = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['chunks']:
                self._debug['chunks']['arr'] = []
            self._debug['chunks']['arr'].append({'start': self._io.pos()})
            _t_chunks = GltfBinary.Chunk(self._io, self, self._root)
            _t_chunks._read()
            self.chunks.append(_t_chunks)
            self._debug['chunks']['arr'][len(self.chunks) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['chunks']['end'] = self._io.pos()

    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "version", "length"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x67\x6C\x54\x46":
                raise kaitaistruct.ValidationNotEqualError(b"\x67\x6C\x54\x46", self.magic, self._io, u"/types/header/seq/0")
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u4le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u4le()
            self._debug['length']['end'] = self._io.pos()


    class Chunk(KaitaiStruct):
        SEQ_FIELDS = ["len_data", "type", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u4le()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(GltfBinary.ChunkType, self._io.read_u4le())
            self._debug['type']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            _on = self.type
            if _on == GltfBinary.ChunkType.json:
                self._raw_data = self._io.read_bytes(self.len_data)
                _io__raw_data = KaitaiStream(BytesIO(self._raw_data))
                self.data = GltfBinary.Json(_io__raw_data, self, self._root)
                self.data._read()
            elif _on == GltfBinary.ChunkType.bin:
                self._raw_data = self._io.read_bytes(self.len_data)
                _io__raw_data = KaitaiStream(BytesIO(self._raw_data))
                self.data = GltfBinary.Bin(_io__raw_data, self, self._root)
                self.data._read()
            else:
                self.data = self._io.read_bytes(self.len_data)
            self._debug['data']['end'] = self._io.pos()


    class Json(KaitaiStruct):
        SEQ_FIELDS = ["data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['data']['start'] = self._io.pos()
            self.data = (self._io.read_bytes_full()).decode(u"UTF-8")
            self._debug['data']['end'] = self._io.pos()


    class Bin(KaitaiStruct):
        SEQ_FIELDS = ["data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes_full()
            self._debug['data']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class MicrosoftPe(KaitaiStruct):
    """
    .. seealso::
       Source - https://docs.microsoft.com/en-us/windows/win32/debug/pe-format
    """

    class PeFormat(Enum):
        rom_image = 263
        pe32 = 267
        pe32_plus = 523
    SEQ_FIELDS = ["mz"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['mz']['start'] = self._io.pos()
        self.mz = MicrosoftPe.MzPlaceholder(self._io, self, self._root)
        self.mz._read()
        self._debug['mz']['end'] = self._io.pos()

    class CertificateEntry(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.microsoft.com/en-us/windows/desktop/debug/pe-format#the-attribute-certificate-table-image-only
        """

        class CertificateRevision(Enum):
            revision_1_0 = 256
            revision_2_0 = 512

        class CertificateType(Enum):
            x509 = 1
            pkcs_signed_data = 2
            reserved_1 = 3
            ts_stack_signed = 4
        SEQ_FIELDS = ["length", "revision", "certificate_type", "certificate_bytes"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u4le()
            self._debug['length']['end'] = self._io.pos()
            self._debug['revision']['start'] = self._io.pos()
            self.revision = KaitaiStream.resolve_enum(MicrosoftPe.CertificateEntry.CertificateRevision, self._io.read_u2le())
            self._debug['revision']['end'] = self._io.pos()
            self._debug['certificate_type']['start'] = self._io.pos()
            self.certificate_type = KaitaiStream.resolve_enum(MicrosoftPe.CertificateEntry.CertificateType, self._io.read_u2le())
            self._debug['certificate_type']['end'] = self._io.pos()
            self._debug['certificate_bytes']['start'] = self._io.pos()
            self.certificate_bytes = self._io.read_bytes((self.length - 8))
            self._debug['certificate_bytes']['end'] = self._io.pos()


    class OptionalHeaderWindows(KaitaiStruct):

        class SubsystemEnum(Enum):
            unknown = 0
            native = 1
            windows_gui = 2
            windows_cui = 3
            posix_cui = 7
            windows_ce_gui = 9
            efi_application = 10
            efi_boot_service_driver = 11
            efi_runtime_driver = 12
            efi_rom = 13
            xbox = 14
            windows_boot_application = 16
        SEQ_FIELDS = ["image_base_32", "image_base_64", "section_alignment", "file_alignment", "major_operating_system_version", "minor_operating_system_version", "major_image_version", "minor_image_version", "major_subsystem_version", "minor_subsystem_version", "win32_version_value", "size_of_image", "size_of_headers", "check_sum", "subsystem", "dll_characteristics", "size_of_stack_reserve_32", "size_of_stack_reserve_64", "size_of_stack_commit_32", "size_of_stack_commit_64", "size_of_heap_reserve_32", "size_of_heap_reserve_64", "size_of_heap_commit_32", "size_of_heap_commit_64", "loader_flags", "number_of_rva_and_sizes"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if self._parent.std.format == MicrosoftPe.PeFormat.pe32:
                self._debug['image_base_32']['start'] = self._io.pos()
                self.image_base_32 = self._io.read_u4le()
                self._debug['image_base_32']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32_plus:
                self._debug['image_base_64']['start'] = self._io.pos()
                self.image_base_64 = self._io.read_u8le()
                self._debug['image_base_64']['end'] = self._io.pos()

            self._debug['section_alignment']['start'] = self._io.pos()
            self.section_alignment = self._io.read_u4le()
            self._debug['section_alignment']['end'] = self._io.pos()
            self._debug['file_alignment']['start'] = self._io.pos()
            self.file_alignment = self._io.read_u4le()
            self._debug['file_alignment']['end'] = self._io.pos()
            self._debug['major_operating_system_version']['start'] = self._io.pos()
            self.major_operating_system_version = self._io.read_u2le()
            self._debug['major_operating_system_version']['end'] = self._io.pos()
            self._debug['minor_operating_system_version']['start'] = self._io.pos()
            self.minor_operating_system_version = self._io.read_u2le()
            self._debug['minor_operating_system_version']['end'] = self._io.pos()
            self._debug['major_image_version']['start'] = self._io.pos()
            self.major_image_version = self._io.read_u2le()
            self._debug['major_image_version']['end'] = self._io.pos()
            self._debug['minor_image_version']['start'] = self._io.pos()
            self.minor_image_version = self._io.read_u2le()
            self._debug['minor_image_version']['end'] = self._io.pos()
            self._debug['major_subsystem_version']['start'] = self._io.pos()
            self.major_subsystem_version = self._io.read_u2le()
            self._debug['major_subsystem_version']['end'] = self._io.pos()
            self._debug['minor_subsystem_version']['start'] = self._io.pos()
            self.minor_subsystem_version = self._io.read_u2le()
            self._debug['minor_subsystem_version']['end'] = self._io.pos()
            self._debug['win32_version_value']['start'] = self._io.pos()
            self.win32_version_value = self._io.read_u4le()
            self._debug['win32_version_value']['end'] = self._io.pos()
            self._debug['size_of_image']['start'] = self._io.pos()
            self.size_of_image = self._io.read_u4le()
            self._debug['size_of_image']['end'] = self._io.pos()
            self._debug['size_of_headers']['start'] = self._io.pos()
            self.size_of_headers = self._io.read_u4le()
            self._debug['size_of_headers']['end'] = self._io.pos()
            self._debug['check_sum']['start'] = self._io.pos()
            self.check_sum = self._io.read_u4le()
            self._debug['check_sum']['end'] = self._io.pos()
            self._debug['subsystem']['start'] = self._io.pos()
            self.subsystem = KaitaiStream.resolve_enum(MicrosoftPe.OptionalHeaderWindows.SubsystemEnum, self._io.read_u2le())
            self._debug['subsystem']['end'] = self._io.pos()
            self._debug['dll_characteristics']['start'] = self._io.pos()
            self.dll_characteristics = self._io.read_u2le()
            self._debug['dll_characteristics']['end'] = self._io.pos()
            if self._parent.std.format == MicrosoftPe.PeFormat.pe32:
                self._debug['size_of_stack_reserve_32']['start'] = self._io.pos()
                self.size_of_stack_reserve_32 = self._io.read_u4le()
                self._debug['size_of_stack_reserve_32']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32_plus:
                self._debug['size_of_stack_reserve_64']['start'] = self._io.pos()
                self.size_of_stack_reserve_64 = self._io.read_u8le()
                self._debug['size_of_stack_reserve_64']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32:
                self._debug['size_of_stack_commit_32']['start'] = self._io.pos()
                self.size_of_stack_commit_32 = self._io.read_u4le()
                self._debug['size_of_stack_commit_32']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32_plus:
                self._debug['size_of_stack_commit_64']['start'] = self._io.pos()
                self.size_of_stack_commit_64 = self._io.read_u8le()
                self._debug['size_of_stack_commit_64']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32:
                self._debug['size_of_heap_reserve_32']['start'] = self._io.pos()
                self.size_of_heap_reserve_32 = self._io.read_u4le()
                self._debug['size_of_heap_reserve_32']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32_plus:
                self._debug['size_of_heap_reserve_64']['start'] = self._io.pos()
                self.size_of_heap_reserve_64 = self._io.read_u8le()
                self._debug['size_of_heap_reserve_64']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32:
                self._debug['size_of_heap_commit_32']['start'] = self._io.pos()
                self.size_of_heap_commit_32 = self._io.read_u4le()
                self._debug['size_of_heap_commit_32']['end'] = self._io.pos()

            if self._parent.std.format == MicrosoftPe.PeFormat.pe32_plus:
                self._debug['size_of_heap_commit_64']['start'] = self._io.pos()
                self.size_of_heap_commit_64 = self._io.read_u8le()
                self._debug['size_of_heap_commit_64']['end'] = self._io.pos()

            self._debug['loader_flags']['start'] = self._io.pos()
            self.loader_flags = self._io.read_u4le()
            self._debug['loader_flags']['end'] = self._io.pos()
            self._debug['number_of_rva_and_sizes']['start'] = self._io.pos()
            self.number_of_rva_and_sizes = self._io.read_u4le()
            self._debug['number_of_rva_and_sizes']['end'] = self._io.pos()


    class OptionalHeaderDataDirs(KaitaiStruct):
        SEQ_FIELDS = ["export_table", "import_table", "resource_table", "exception_table", "certificate_table", "base_relocation_table", "debug", "architecture", "global_ptr", "tls_table", "load_config_table", "bound_import", "iat", "delay_import_descriptor", "clr_runtime_header"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['export_table']['start'] = self._io.pos()
            self.export_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.export_table._read()
            self._debug['export_table']['end'] = self._io.pos()
            self._debug['import_table']['start'] = self._io.pos()
            self.import_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.import_table._read()
            self._debug['import_table']['end'] = self._io.pos()
            self._debug['resource_table']['start'] = self._io.pos()
            self.resource_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.resource_table._read()
            self._debug['resource_table']['end'] = self._io.pos()
            self._debug['exception_table']['start'] = self._io.pos()
            self.exception_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.exception_table._read()
            self._debug['exception_table']['end'] = self._io.pos()
            self._debug['certificate_table']['start'] = self._io.pos()
            self.certificate_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.certificate_table._read()
            self._debug['certificate_table']['end'] = self._io.pos()
            self._debug['base_relocation_table']['start'] = self._io.pos()
            self.base_relocation_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.base_relocation_table._read()
            self._debug['base_relocation_table']['end'] = self._io.pos()
            self._debug['debug']['start'] = self._io.pos()
            self.debug = MicrosoftPe.DataDir(self._io, self, self._root)
            self.debug._read()
            self._debug['debug']['end'] = self._io.pos()
            self._debug['architecture']['start'] = self._io.pos()
            self.architecture = MicrosoftPe.DataDir(self._io, self, self._root)
            self.architecture._read()
            self._debug['architecture']['end'] = self._io.pos()
            self._debug['global_ptr']['start'] = self._io.pos()
            self.global_ptr = MicrosoftPe.DataDir(self._io, self, self._root)
            self.global_ptr._read()
            self._debug['global_ptr']['end'] = self._io.pos()
            self._debug['tls_table']['start'] = self._io.pos()
            self.tls_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.tls_table._read()
            self._debug['tls_table']['end'] = self._io.pos()
            self._debug['load_config_table']['start'] = self._io.pos()
            self.load_config_table = MicrosoftPe.DataDir(self._io, self, self._root)
            self.load_config_table._read()
            self._debug['load_config_table']['end'] = self._io.pos()
            self._debug['bound_import']['start'] = self._io.pos()
            self.bound_import = MicrosoftPe.DataDir(self._io, self, self._root)
            self.bound_import._read()
            self._debug['bound_import']['end'] = self._io.pos()
            self._debug['iat']['start'] = self._io.pos()
            self.iat = MicrosoftPe.DataDir(self._io, self, self._root)
            self.iat._read()
            self._debug['iat']['end'] = self._io.pos()
            self._debug['delay_import_descriptor']['start'] = self._io.pos()
            self.delay_import_descriptor = MicrosoftPe.DataDir(self._io, self, self._root)
            self.delay_import_descriptor._read()
            self._debug['delay_import_descriptor']['end'] = self._io.pos()
            self._debug['clr_runtime_header']['start'] = self._io.pos()
            self.clr_runtime_header = MicrosoftPe.DataDir(self._io, self, self._root)
            self.clr_runtime_header._read()
            self._debug['clr_runtime_header']['end'] = self._io.pos()


    class DataDir(KaitaiStruct):
        SEQ_FIELDS = ["virtual_address", "size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['virtual_address']['start'] = self._io.pos()
            self.virtual_address = self._io.read_u4le()
            self._debug['virtual_address']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()


    class CoffSymbol(KaitaiStruct):
        SEQ_FIELDS = ["name_annoying", "value", "section_number", "type", "storage_class", "number_of_aux_symbols"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name_annoying']['start'] = self._io.pos()
            self._raw_name_annoying = self._io.read_bytes(8)
            _io__raw_name_annoying = KaitaiStream(BytesIO(self._raw_name_annoying))
            self.name_annoying = MicrosoftPe.Annoyingstring(_io__raw_name_annoying, self, self._root)
            self.name_annoying._read()
            self._debug['name_annoying']['end'] = self._io.pos()
            self._debug['value']['start'] = self._io.pos()
            self.value = self._io.read_u4le()
            self._debug['value']['end'] = self._io.pos()
            self._debug['section_number']['start'] = self._io.pos()
            self.section_number = self._io.read_u2le()
            self._debug['section_number']['end'] = self._io.pos()
            self._debug['type']['start'] = self._io.pos()
            self.type = self._io.read_u2le()
            self._debug['type']['end'] = self._io.pos()
            self._debug['storage_class']['start'] = self._io.pos()
            self.storage_class = self._io.read_u1()
            self._debug['storage_class']['end'] = self._io.pos()
            self._debug['number_of_aux_symbols']['start'] = self._io.pos()
            self.number_of_aux_symbols = self._io.read_u1()
            self._debug['number_of_aux_symbols']['end'] = self._io.pos()

        @property
        def section(self):
            if hasattr(self, '_m_section'):
                return self._m_section if hasattr(self, '_m_section') else None

            self._m_section = self._root.pe.sections[(self.section_number - 1)]
            return self._m_section if hasattr(self, '_m_section') else None

        @property
        def data(self):
            if hasattr(self, '_m_data'):
                return self._m_data if hasattr(self, '_m_data') else None

            _pos = self._io.pos()
            self._io.seek((self.section.pointer_to_raw_data + self.value))
            self._debug['_m_data']['start'] = self._io.pos()
            self._m_data = self._io.read_bytes(1)
            self._debug['_m_data']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_data if hasattr(self, '_m_data') else None


    class PeHeader(KaitaiStruct):
        SEQ_FIELDS = ["pe_signature", "coff_hdr", "optional_hdr", "sections"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['pe_signature']['start'] = self._io.pos()
            self.pe_signature = self._io.read_bytes(4)
            self._debug['pe_signature']['end'] = self._io.pos()
            if not self.pe_signature == b"\x50\x45\x00\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x50\x45\x00\x00", self.pe_signature, self._io, u"/types/pe_header/seq/0")
            self._debug['coff_hdr']['start'] = self._io.pos()
            self.coff_hdr = MicrosoftPe.CoffHeader(self._io, self, self._root)
            self.coff_hdr._read()
            self._debug['coff_hdr']['end'] = self._io.pos()
            self._debug['optional_hdr']['start'] = self._io.pos()
            self._raw_optional_hdr = self._io.read_bytes(self.coff_hdr.size_of_optional_header)
            _io__raw_optional_hdr = KaitaiStream(BytesIO(self._raw_optional_hdr))
            self.optional_hdr = MicrosoftPe.OptionalHeader(_io__raw_optional_hdr, self, self._root)
            self.optional_hdr._read()
            self._debug['optional_hdr']['end'] = self._io.pos()
            self._debug['sections']['start'] = self._io.pos()
            self.sections = [None] * (self.coff_hdr.number_of_sections)
            for i in range(self.coff_hdr.number_of_sections):
                if not 'arr' in self._debug['sections']:
                    self._debug['sections']['arr'] = []
                self._debug['sections']['arr'].append({'start': self._io.pos()})
                _t_sections = MicrosoftPe.Section(self._io, self, self._root)
                _t_sections._read()
                self.sections[i] = _t_sections
                self._debug['sections']['arr'][i]['end'] = self._io.pos()

            self._debug['sections']['end'] = self._io.pos()

        @property
        def certificate_table(self):
            if hasattr(self, '_m_certificate_table'):
                return self._m_certificate_table if hasattr(self, '_m_certificate_table') else None

            if self.optional_hdr.data_dirs.certificate_table.virtual_address != 0:
                _pos = self._io.pos()
                self._io.seek(self.optional_hdr.data_dirs.certificate_table.virtual_address)
                self._debug['_m_certificate_table']['start'] = self._io.pos()
                self._raw__m_certificate_table = self._io.read_bytes(self.optional_hdr.data_dirs.certificate_table.size)
                _io__raw__m_certificate_table = KaitaiStream(BytesIO(self._raw__m_certificate_table))
                self._m_certificate_table = MicrosoftPe.CertificateTable(_io__raw__m_certificate_table, self, self._root)
                self._m_certificate_table._read()
                self._debug['_m_certificate_table']['end'] = self._io.pos()
                self._io.seek(_pos)

            return self._m_certificate_table if hasattr(self, '_m_certificate_table') else None


    class OptionalHeader(KaitaiStruct):
        SEQ_FIELDS = ["std", "windows", "data_dirs"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['std']['start'] = self._io.pos()
            self.std = MicrosoftPe.OptionalHeaderStd(self._io, self, self._root)
            self.std._read()
            self._debug['std']['end'] = self._io.pos()
            self._debug['windows']['start'] = self._io.pos()
            self.windows = MicrosoftPe.OptionalHeaderWindows(self._io, self, self._root)
            self.windows._read()
            self._debug['windows']['end'] = self._io.pos()
            self._debug['data_dirs']['start'] = self._io.pos()
            self.data_dirs = MicrosoftPe.OptionalHeaderDataDirs(self._io, self, self._root)
            self.data_dirs._read()
            self._debug['data_dirs']['end'] = self._io.pos()


    class Section(KaitaiStruct):
        SEQ_FIELDS = ["name", "virtual_size", "virtual_address", "size_of_raw_data", "pointer_to_raw_data", "pointer_to_relocations", "pointer_to_linenumbers", "number_of_relocations", "number_of_linenumbers", "characteristics"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name']['start'] = self._io.pos()
            self.name = (KaitaiStream.bytes_strip_right(self._io.read_bytes(8), 0)).decode(u"UTF-8")
            self._debug['name']['end'] = self._io.pos()
            self._debug['virtual_size']['start'] = self._io.pos()
            self.virtual_size = self._io.read_u4le()
            self._debug['virtual_size']['end'] = self._io.pos()
            self._debug['virtual_address']['start'] = self._io.pos()
            self.virtual_address = self._io.read_u4le()
            self._debug['virtual_address']['end'] = self._io.pos()
            self._debug['size_of_raw_data']['start'] = self._io.pos()
            self.size_of_raw_data = self._io.read_u4le()
            self._debug['size_of_raw_data']['end'] = self._io.pos()
            self._debug['pointer_to_raw_data']['start'] = self._io.pos()
            self.pointer_to_raw_data = self._io.read_u4le()
            self._debug['pointer_to_raw_data']['end'] = self._io.pos()
            self._debug['pointer_to_relocations']['start'] = self._io.pos()
            self.pointer_to_relocations = self._io.read_u4le()
            self._debug['pointer_to_relocations']['end'] = self._io.pos()
            self._debug['pointer_to_linenumbers']['start'] = self._io.pos()
            self.pointer_to_linenumbers = self._io.read_u4le()
            self._debug['pointer_to_linenumbers']['end'] = self._io.pos()
            self._debug['number_of_relocations']['start'] = self._io.pos()
            self.number_of_relocations = self._io.read_u2le()
            self._debug['number_of_relocations']['end'] = self._io.pos()
            self._debug['number_of_linenumbers']['start'] = self._io.pos()
            self.number_of_linenumbers = self._io.read_u2le()
            self._debug['number_of_linenumbers']['end'] = self._io.pos()
            self._debug['characteristics']['start'] = self._io.pos()
            self.characteristics = self._io.read_u4le()
            self._debug['characteristics']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek(self.pointer_to_raw_data)
            self._debug['_m_body']['start'] = self._io.pos()
            self._m_body = self._io.read_bytes(self.size_of_raw_data)
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class CertificateTable(KaitaiStruct):
        SEQ_FIELDS = ["items"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['items']['start'] = self._io.pos()
            self.items = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['items']:
                    self._debug['items']['arr'] = []
                self._debug['items']['arr'].append({'start': self._io.pos()})
                _t_items = MicrosoftPe.CertificateEntry(self._io, self, self._root)
                _t_items._read()
                self.items.append(_t_items)
                self._debug['items']['arr'][len(self.items) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['items']['end'] = self._io.pos()


    class MzPlaceholder(KaitaiStruct):
        SEQ_FIELDS = ["magic", "data1", "ofs_pe"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(2)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x4D\x5A":
                raise kaitaistruct.ValidationNotEqualError(b"\x4D\x5A", self.magic, self._io, u"/types/mz_placeholder/seq/0")
            self._debug['data1']['start'] = self._io.pos()
            self.data1 = self._io.read_bytes(58)
            self._debug['data1']['end'] = self._io.pos()
            self._debug['ofs_pe']['start'] = self._io.pos()
            self.ofs_pe = self._io.read_u4le()
            self._debug['ofs_pe']['end'] = self._io.pos()


    class OptionalHeaderStd(KaitaiStruct):
        SEQ_FIELDS = ["format", "major_linker_version", "minor_linker_version", "size_of_code", "size_of_initialized_data", "size_of_uninitialized_data", "address_of_entry_point", "base_of_code", "base_of_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['format']['start'] = self._io.pos()
            self.format = KaitaiStream.resolve_enum(MicrosoftPe.PeFormat, self._io.read_u2le())
            self._debug['format']['end'] = self._io.pos()
            self._debug['major_linker_version']['start'] = self._io.pos()
            self.major_linker_version = self._io.read_u1()
            self._debug['major_linker_version']['end'] = self._io.pos()
            self._debug['minor_linker_version']['start'] = self._io.pos()
            self.minor_linker_version = self._io.read_u1()
            self._debug['minor_linker_version']['end'] = self._io.pos()
            self._debug['size_of_code']['start'] = self._io.pos()
            self.size_of_code = self._io.read_u4le()
            self._debug['size_of_code']['end'] = self._io.pos()
            self._debug['size_of_initialized_data']['start'] = self._io.pos()
            self.size_of_initialized_data = self._io.read_u4le()
            self._debug['size_of_initialized_data']['end'] = self._io.pos()
            self._debug['size_of_uninitialized_data']['start'] = self._io.pos()
            self.size_of_uninitialized_data = self._io.read_u4le()
            self._debug['size_of_uninitialized_data']['end'] = self._io.pos()
            self._debug['address_of_entry_point']['start'] = self._io.pos()
            self.address_of_entry_point = self._io.read_u4le()
            self._debug['address_of_entry_point']['end'] = self._io.pos()
            self._debug['base_of_code']['start'] = self._io.pos()
            self.base_of_code = self._io.read_u4le()
            self._debug['base_of_code']['end'] = self._io.pos()
            if self.format == MicrosoftPe.PeFormat.pe32:
                self._debug['base_of_data']['start'] = self._io.pos()
                self.base_of_data = self._io.read_u4le()
                self._debug['base_of_data']['end'] = self._io.pos()



    class CoffHeader(KaitaiStruct):
        """
        .. seealso::
           3.3. COFF File Header (Object and Image)
        """

        class MachineType(Enum):
            unknown = 0
            i386 = 332
            r4000 = 358
            wcemipsv2 = 361
            alpha = 388
            sh3 = 418
            sh3dsp = 419
            sh4 = 422
            sh5 = 424
            arm = 448
            thumb = 450
            armnt = 452
            am33 = 467
            powerpc = 496
            powerpcfp = 497
            ia64 = 512
            mips16 = 614
            mipsfpu = 870
            mipsfpu16 = 1126
            ebc = 3772
            riscv32 = 20530
            riscv64 = 20580
            riscv128 = 20776
            amd64 = 34404
            m32r = 36929
            arm64 = 43620
        SEQ_FIELDS = ["machine", "number_of_sections", "time_date_stamp", "pointer_to_symbol_table", "number_of_symbols", "size_of_optional_header", "characteristics"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['machine']['start'] = self._io.pos()
            self.machine = KaitaiStream.resolve_enum(MicrosoftPe.CoffHeader.MachineType, self._io.read_u2le())
            self._debug['machine']['end'] = self._io.pos()
            self._debug['number_of_sections']['start'] = self._io.pos()
            self.number_of_sections = self._io.read_u2le()
            self._debug['number_of_sections']['end'] = self._io.pos()
            self._debug['time_date_stamp']['start'] = self._io.pos()
            self.time_date_stamp = self._io.read_u4le()
            self._debug['time_date_stamp']['end'] = self._io.pos()
            self._debug['pointer_to_symbol_table']['start'] = self._io.pos()
            self.pointer_to_symbol_table = self._io.read_u4le()
            self._debug['pointer_to_symbol_table']['end'] = self._io.pos()
            self._debug['number_of_symbols']['start'] = self._io.pos()
            self.number_of_symbols = self._io.read_u4le()
            self._debug['number_of_symbols']['end'] = self._io.pos()
            self._debug['size_of_optional_header']['start'] = self._io.pos()
            self.size_of_optional_header = self._io.read_u2le()
            self._debug['size_of_optional_header']['end'] = self._io.pos()
            self._debug['characteristics']['start'] = self._io.pos()
            self.characteristics = self._io.read_u2le()
            self._debug['characteristics']['end'] = self._io.pos()

        @property
        def symbol_table_size(self):
            if hasattr(self, '_m_symbol_table_size'):
                return self._m_symbol_table_size if hasattr(self, '_m_symbol_table_size') else None

            self._m_symbol_table_size = (self.number_of_symbols * 18)
            return self._m_symbol_table_size if hasattr(self, '_m_symbol_table_size') else None

        @property
        def symbol_name_table_offset(self):
            if hasattr(self, '_m_symbol_name_table_offset'):
                return self._m_symbol_name_table_offset if hasattr(self, '_m_symbol_name_table_offset') else None

            self._m_symbol_name_table_offset = (self.pointer_to_symbol_table + self.symbol_table_size)
            return self._m_symbol_name_table_offset if hasattr(self, '_m_symbol_name_table_offset') else None

        @property
        def symbol_name_table_size(self):
            if hasattr(self, '_m_symbol_name_table_size'):
                return self._m_symbol_name_table_size if hasattr(self, '_m_symbol_name_table_size') else None

            _pos = self._io.pos()
            self._io.seek(self.symbol_name_table_offset)
            self._debug['_m_symbol_name_table_size']['start'] = self._io.pos()
            self._m_symbol_name_table_size = self._io.read_u4le()
            self._debug['_m_symbol_name_table_size']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_symbol_name_table_size if hasattr(self, '_m_symbol_name_table_size') else None

        @property
        def symbol_table(self):
            if hasattr(self, '_m_symbol_table'):
                return self._m_symbol_table if hasattr(self, '_m_symbol_table') else None

            _pos = self._io.pos()
            self._io.seek(self.pointer_to_symbol_table)
            self._debug['_m_symbol_table']['start'] = self._io.pos()
            self._m_symbol_table = [None] * (self.number_of_symbols)
            for i in range(self.number_of_symbols):
                if not 'arr' in self._debug['_m_symbol_table']:
                    self._debug['_m_symbol_table']['arr'] = []
                self._debug['_m_symbol_table']['arr'].append({'start': self._io.pos()})
                _t__m_symbol_table = MicrosoftPe.CoffSymbol(self._io, self, self._root)
                _t__m_symbol_table._read()
                self._m_symbol_table[i] = _t__m_symbol_table
                self._debug['_m_symbol_table']['arr'][i]['end'] = self._io.pos()

            self._debug['_m_symbol_table']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_symbol_table if hasattr(self, '_m_symbol_table') else None


    class Annoyingstring(KaitaiStruct):
        SEQ_FIELDS = []
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            pass

        @property
        def name_from_offset(self):
            if hasattr(self, '_m_name_from_offset'):
                return self._m_name_from_offset if hasattr(self, '_m_name_from_offset') else None

            if self.name_zeroes == 0:
                io = self._root._io
                _pos = io.pos()
                io.seek(((self._parent._parent.symbol_name_table_offset + self.name_offset) if self.name_zeroes == 0 else 0))
                self._debug['_m_name_from_offset']['start'] = io.pos()
                self._m_name_from_offset = (io.read_bytes_term(0, False, True, False)).decode(u"ascii")
                self._debug['_m_name_from_offset']['end'] = io.pos()
                io.seek(_pos)

            return self._m_name_from_offset if hasattr(self, '_m_name_from_offset') else None

        @property
        def name_offset(self):
            if hasattr(self, '_m_name_offset'):
                return self._m_name_offset if hasattr(self, '_m_name_offset') else None

            _pos = self._io.pos()
            self._io.seek(4)
            self._debug['_m_name_offset']['start'] = self._io.pos()
            self._m_name_offset = self._io.read_u4le()
            self._debug['_m_name_offset']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_name_offset if hasattr(self, '_m_name_offset') else None

        @property
        def name(self):
            if hasattr(self, '_m_name'):
                return self._m_name if hasattr(self, '_m_name') else None

            self._m_name = (self.name_from_offset if self.name_zeroes == 0 else self.name_from_short)
            return self._m_name if hasattr(self, '_m_name') else None

        @property
        def name_zeroes(self):
            if hasattr(self, '_m_name_zeroes'):
                return self._m_name_zeroes if hasattr(self, '_m_name_zeroes') else None

            _pos = self._io.pos()
            self._io.seek(0)
            self._debug['_m_name_zeroes']['start'] = self._io.pos()
            self._m_name_zeroes = self._io.read_u4le()
            self._debug['_m_name_zeroes']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_name_zeroes if hasattr(self, '_m_name_zeroes') else None

        @property
        def name_from_short(self):
            if hasattr(self, '_m_name_from_short'):
                return self._m_name_from_short if hasattr(self, '_m_name_from_short') else None

            if self.name_zeroes != 0:
                _pos = self._io.pos()
                self._io.seek(0)
                self._debug['_m_name_from_short']['start'] = self._io.pos()
                self._m_name_from_short = (self._io.read_bytes_term(0, False, True, False)).decode(u"ascii")
                self._debug['_m_name_from_short']['end'] = self._io.pos()
                self._io.seek(_pos)

            return self._m_name_from_short if hasattr(self, '_m_name_from_short') else None


    @property
    def pe(self):
        if hasattr(self, '_m_pe'):
            return self._m_pe if hasattr(self, '_m_pe') else None

        _pos = self._io.pos()
        self._io.seek(self.mz.ofs_pe)
        self._debug['_m_pe']['start'] = self._io.pos()
        self._m_pe = MicrosoftPe.PeHeader(self._io, self, self._root)
        self._m_pe._read()
        self._debug['_m_pe']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_pe if hasattr(self, '_m_pe') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class TlsClientHello(KaitaiStruct):
    SEQ_FIELDS = ["version", "random", "session_id", "cipher_suites", "compression_methods", "extensions"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['version']['start'] = self._io.pos()
        self.version = TlsClientHello.Version(self._io, self, self._root)
        self.version._read()
        self._debug['version']['end'] = self._io.pos()
        self._debug['random']['start'] = self._io.pos()
        self.random = TlsClientHello.Random(self._io, self, self._root)
        self.random._read()
        self._debug['random']['end'] = self._io.pos()
        self._debug['session_id']['start'] = self._io.pos()
        self.session_id = TlsClientHello.SessionId(self._io, self, self._root)
        self.session_id._read()
        self._debug['session_id']['end'] = self._io.pos()
        self._debug['cipher_suites']['start'] = self._io.pos()
        self.cipher_suites = TlsClientHello.CipherSuites(self._io, self, self._root)
        self.cipher_suites._read()
        self._debug['cipher_suites']['end'] = self._io.pos()
        self._debug['compression_methods']['start'] = self._io.pos()
        self.compression_methods = TlsClientHello.CompressionMethods(self._io, self, self._root)
        self.compression_methods._read()
        self._debug['compression_methods']['end'] = self._io.pos()
        if self._io.is_eof() == False:
            self._debug['extensions']['start'] = self._io.pos()
            self.extensions = TlsClientHello.Extensions(self._io, self, self._root)
            self.extensions._read()
            self._debug['extensions']['end'] = self._io.pos()


    class ServerName(KaitaiStruct):
        SEQ_FIELDS = ["name_type", "length", "host_name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name_type']['start'] = self._io.pos()
            self.name_type = self._io.read_u1()
            self._debug['name_type']['end'] = self._io.pos()
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u2be()
            self._debug['length']['end'] = self._io.pos()
            self._debug['host_name']['start'] = self._io.pos()
            self.host_name = self._io.read_bytes(self.length)
            self._debug['host_name']['end'] = self._io.pos()


    class Random(KaitaiStruct):
        SEQ_FIELDS = ["gmt_unix_time", "random"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['gmt_unix_time']['start'] = self._io.pos()
            self.gmt_unix_time = self._io.read_u4be()
            self._debug['gmt_unix_time']['end'] = self._io.pos()
            self._debug['random']['start'] = self._io.pos()
            self.random = self._io.read_bytes(28)
            self._debug['random']['end'] = self._io.pos()


    class SessionId(KaitaiStruct):
        SEQ_FIELDS = ["len", "sid"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u1()
            self._debug['len']['end'] = self._io.pos()
            self._debug['sid']['start'] = self._io.pos()
            self.sid = self._io.read_bytes(self.len)
            self._debug['sid']['end'] = self._io.pos()


    class Sni(KaitaiStruct):
        SEQ_FIELDS = ["list_length", "server_names"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['list_length']['start'] = self._io.pos()
            self.list_length = self._io.read_u2be()
            self._debug['list_length']['end'] = self._io.pos()
            self._debug['server_names']['start'] = self._io.pos()
            self.server_names = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['server_names']:
                    self._debug['server_names']['arr'] = []
                self._debug['server_names']['arr'].append({'start': self._io.pos()})
                _t_server_names = TlsClientHello.ServerName(self._io, self, self._root)
                _t_server_names._read()
                self.server_names.append(_t_server_names)
                self._debug['server_names']['arr'][len(self.server_names) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['server_names']['end'] = self._io.pos()


    class CipherSuites(KaitaiStruct):
        SEQ_FIELDS = ["len", "cipher_suites"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u2be()
            self._debug['len']['end'] = self._io.pos()
            self._debug['cipher_suites']['start'] = self._io.pos()
            self.cipher_suites = [None] * (self.len // 2)
            for i in range(self.len // 2):
                if not 'arr' in self._debug['cipher_suites']:
                    self._debug['cipher_suites']['arr'] = []
                self._debug['cipher_suites']['arr'].append({'start': self._io.pos()})
                self.cipher_suites[i] = self._io.read_u2be()
                self._debug['cipher_suites']['arr'][i]['end'] = self._io.pos()

            self._debug['cipher_suites']['end'] = self._io.pos()


    class CompressionMethods(KaitaiStruct):
        SEQ_FIELDS = ["len", "compression_methods"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u1()
            self._debug['len']['end'] = self._io.pos()
            self._debug['compression_methods']['start'] = self._io.pos()
            self.compression_methods = self._io.read_bytes(self.len)
            self._debug['compression_methods']['end'] = self._io.pos()


    class Alpn(KaitaiStruct):
        SEQ_FIELDS = ["ext_len", "alpn_protocols"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ext_len']['start'] = self._io.pos()
            self.ext_len = self._io.read_u2be()
            self._debug['ext_len']['end'] = self._io.pos()
            self._debug['alpn_protocols']['start'] = self._io.pos()
            self.alpn_protocols = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['alpn_protocols']:
                    self._debug['alpn_protocols']['arr'] = []
                self._debug['alpn_protocols']['arr'].append({'start': self._io.pos()})
                _t_alpn_protocols = TlsClientHello.Protocol(self._io, self, self._root)
                _t_alpn_protocols._read()
                self.alpn_protocols.append(_t_alpn_protocols)
                self._debug['alpn_protocols']['arr'][len(self.alpn_protocols) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['alpn_protocols']['end'] = self._io.pos()


    class Extensions(KaitaiStruct):
        SEQ_FIELDS = ["len", "extensions"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u2be()
            self._debug['len']['end'] = self._io.pos()
            self._debug['extensions']['start'] = self._io.pos()
            self.extensions = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['extensions']:
                    self._debug['extensions']['arr'] = []
                self._debug['extensions']['arr'].append({'start': self._io.pos()})
                _t_extensions = TlsClientHello.Extension(self._io, self, self._root)
                _t_extensions._read()
                self.extensions.append(_t_extensions)
                self._debug['extensions']['arr'][len(self.extensions) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['extensions']['end'] = self._io.pos()


    class Version(KaitaiStruct):
        SEQ_FIELDS = ["major", "minor"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['major']['start'] = self._io.pos()
            self.major = self._io.read_u1()
            self._debug['major']['end'] = self._io.pos()
            self._debug['minor']['start'] = self._io.pos()
            self.minor = self._io.read_u1()
            self._debug['minor']['end'] = self._io.pos()


    class Protocol(KaitaiStruct):
        SEQ_FIELDS = ["strlen", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['strlen']['start'] = self._io.pos()
            self.strlen = self._io.read_u1()
            self._debug['strlen']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = self._io.read_bytes(self.strlen)
            self._debug['name']['end'] = self._io.pos()


    class Extension(KaitaiStruct):
        SEQ_FIELDS = ["type", "len", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = self._io.read_u2be()
            self._debug['type']['end'] = self._io.pos()
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u2be()
            self._debug['len']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.type
            if _on == 0:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = TlsClientHello.Sni(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == 16:
                self._raw_body = self._io.read_bytes(self.len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = TlsClientHello.Alpn(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.len)
            self._debug['body']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class FalloutDat(KaitaiStruct):

    class Compression(Enum):
        none = 32
        lzss = 64
    SEQ_FIELDS = ["folder_count", "unknown1", "unknown2", "timestamp", "folder_names", "folders"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['folder_count']['start'] = self._io.pos()
        self.folder_count = self._io.read_u4be()
        self._debug['folder_count']['end'] = self._io.pos()
        self._debug['unknown1']['start'] = self._io.pos()
        self.unknown1 = self._io.read_u4be()
        self._debug['unknown1']['end'] = self._io.pos()
        self._debug['unknown2']['start'] = self._io.pos()
        self.unknown2 = self._io.read_u4be()
        self._debug['unknown2']['end'] = self._io.pos()
        self._debug['timestamp']['start'] = self._io.pos()
        self.timestamp = self._io.read_u4be()
        self._debug['timestamp']['end'] = self._io.pos()
        self._debug['folder_names']['start'] = self._io.pos()
        self.folder_names = [None] * (self.folder_count)
        for i in range(self.folder_count):
            if not 'arr' in self._debug['folder_names']:
                self._debug['folder_names']['arr'] = []
            self._debug['folder_names']['arr'].append({'start': self._io.pos()})
            _t_folder_names = FalloutDat.Pstr(self._io, self, self._root)
            _t_folder_names._read()
            self.folder_names[i] = _t_folder_names
            self._debug['folder_names']['arr'][i]['end'] = self._io.pos()

        self._debug['folder_names']['end'] = self._io.pos()
        self._debug['folders']['start'] = self._io.pos()
        self.folders = [None] * (self.folder_count)
        for i in range(self.folder_count):
            if not 'arr' in self._debug['folders']:
                self._debug['folders']['arr'] = []
            self._debug['folders']['arr'].append({'start': self._io.pos()})
            _t_folders = FalloutDat.Folder(self._io, self, self._root)
            _t_folders._read()
            self.folders[i] = _t_folders
            self._debug['folders']['arr'][i]['end'] = self._io.pos()

        self._debug['folders']['end'] = self._io.pos()

    class Pstr(KaitaiStruct):
        SEQ_FIELDS = ["size", "str"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u1()
            self._debug['size']['end'] = self._io.pos()
            self._debug['str']['start'] = self._io.pos()
            self.str = (self._io.read_bytes(self.size)).decode(u"ASCII")
            self._debug['str']['end'] = self._io.pos()


    class Folder(KaitaiStruct):
        SEQ_FIELDS = ["file_count", "unknown", "flags", "timestamp", "files"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['file_count']['start'] = self._io.pos()
            self.file_count = self._io.read_u4be()
            self._debug['file_count']['end'] = self._io.pos()
            self._debug['unknown']['start'] = self._io.pos()
            self.unknown = self._io.read_u4be()
            self._debug['unknown']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u4be()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4be()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['files']['start'] = self._io.pos()
            self.files = [None] * (self.file_count)
            for i in range(self.file_count):
                if not 'arr' in self._debug['files']:
                    self._debug['files']['arr'] = []
                self._debug['files']['arr'].append({'start': self._io.pos()})
                _t_files = FalloutDat.File(self._io, self, self._root)
                _t_files._read()
                self.files[i] = _t_files
                self._debug['files']['arr'][i]['end'] = self._io.pos()

            self._debug['files']['end'] = self._io.pos()


    class File(KaitaiStruct):
        SEQ_FIELDS = ["name", "flags", "offset", "size_unpacked", "size_packed"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name']['start'] = self._io.pos()
            self.name = FalloutDat.Pstr(self._io, self, self._root)
            self.name._read()
            self._debug['name']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = KaitaiStream.resolve_enum(FalloutDat.Compression, self._io.read_u4be())
            self._debug['flags']['end'] = self._io.pos()
            self._debug['offset']['start'] = self._io.pos()
            self.offset = self._io.read_u4be()
            self._debug['offset']['end'] = self._io.pos()
            self._debug['size_unpacked']['start'] = self._io.pos()
            self.size_unpacked = self._io.read_u4be()
            self._debug['size_unpacked']['end'] = self._io.pos()
            self._debug['size_packed']['start'] = self._io.pos()
            self.size_packed = self._io.read_u4be()
            self._debug['size_packed']['end'] = self._io.pos()

        @property
        def contents(self):
            if hasattr(self, '_m_contents'):
                return self._m_contents if hasattr(self, '_m_contents') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.offset)
            self._debug['_m_contents']['start'] = io.pos()
            self._m_contents = io.read_bytes((self.size_unpacked if self.flags == FalloutDat.Compression.none else self.size_packed))
            self._debug['_m_contents']['end'] = io.pos()
            io.seek(_pos)
            return self._m_contents if hasattr(self, '_m_contents') else None






import zlib
from typing import Dict, Iterator, List, Optional, Type

from . import pdfparser
from .kaitai.parser import KaitaiParser
from .fileutils import Tempfile
from .kaitaimatcher import ast_to_matches
from .logger import getStatusLogger
from .polyfile import Match, Matcher, Submatch, submatcher

log = getStatusLogger("PDF")


def token_length(tok):
    if hasattr(tok, 'token'):
        return len(tok.token)
    else:
        return len(tok[1])


def content_length(content):
    return content[-1].offset.offset - content[0].offset.offset + token_length(content[-1])


def _emit_dict(parsed, parent, pdf_offset):
    dict_obj = Submatch(
        "PDFDictionary",
        '',
        relative_offset=parsed.start.offset.offset - parent.offset + pdf_offset,
        length=parsed.end.offset.offset - parsed.start.offset.offset + len(parsed.end.token),
        parent=parent
    )
    yield dict_obj
    for key, value in parsed:
        if isinstance(value, pdfparser.ParsedDictionary):
            value_end = value.end.offset.offset + len(value.end.token)
        else:
            value_end = value[-1].offset.offset + len(value[-1].token)
        pair_offset = key.offset.offset - dict_obj.offset
        pair = Submatch(
            "KeyValuePair",
            '',
            relative_offset=pair_offset + pdf_offset,
            length=value_end - key.offset.offset,
            parent=dict_obj
        )
        yield pair
        yield Submatch(
            "Key",
            key.token,
            relative_offset=0,
            length=len(key.token),
            parent=pair
        )
        if isinstance(value, pdfparser.ParsedDictionary):
            yield from _emit_dict(value, pair, pdf_offset)
        else:
            value_length = value[-1].offset.offset + len(value[-1].token) - value[0].offset.offset
            yield Submatch(
                "Value",
                ''.join(v.token for v in value),
                relative_offset=value[0].offset.offset - key.offset.offset,
                length=value_length,
                parent=pair
            )


FILTERS_BY_NAME: Dict[str, Type["StreamFilter"]] = {}


class StreamFilter:
    name: str

    def __init__(self, next_decoder: Optional["StreamFilter"] = None):
        self.next_decoder: Optional[StreamFilter] = next_decoder

    def __init_subclass__(cls, **kwargs):
        FILTERS_BY_NAME[f"/{cls.name}Decode"] = cls

    def decode(self, matcher: Matcher, raw_content: bytes, parent: Submatch) -> Iterator[Submatch]:
        raise NotImplementedError()

    def match(self, matcher: Matcher, raw_content: bytes, parent: Submatch) -> Iterator[Submatch]:
        for submatch in self.decode(matcher, raw_content, parent):
            yield submatch
            if submatch.decoded is None:
                if self.next_decoder is not None:
                    log.warning(f"Expected submatch submatch {submatch!r} from decoded by {self.__class__.__name__} "
                                "to have a `decoded` member, but it was `None`")
                continue
            if self.next_decoder is None:
                # recursively match against the deflated contents
                with Tempfile(submatch.decoded) as f:
                    yield from matcher.match(f, parent=submatch)
            else:
                yield from self.next_decoder.match(matcher, submatch.decoded, submatch)


class FlateDecoder(StreamFilter):
    name = "Flate"

    def decode(self, matcher: Matcher, raw_content: bytes, parent: Submatch) -> Iterator[Submatch]:
        try:
            decoded = zlib.decompress(raw_content)
            yield Submatch(
                f"{self.name}Encoded",
                raw_content,
                relative_offset=0,
                length=len(raw_content),
                parent=parent,
                decoded=decoded
            )
        except zlib.error:
            log.warn(f"DEFLATE decoding error near offset {parent.offset}")


class DCTDecoder(StreamFilter):
    name = "DCT"

    def decode(self, matcher: Matcher, raw_content: bytes, parent: Submatch) -> Iterator[Submatch]:
        if raw_content[:1] != b'\xff':
            return
        # This is most likely a JPEG image
        try:
            ast = KaitaiParser.load("image/jpeg.ksy").parse(raw_content).ast
        except Exception as e:
            log.error(str(e))
            ast = None
        if ast is not None:
            yield from ast_to_matches(ast, parent=parent)


class ASCIIHexDecoder(StreamFilter):
    name = "ASCIIHex"

    def decode(self, matcher: Matcher, raw_content: bytes, parent: Submatch) -> Iterator[Submatch]:
        data = bytearray()
        for byte_str in raw_content.replace(b"\n", b" ").split(b" "):
            byte_str = byte_str.strip()
            if byte_str.endswith(b">"):
                byte_str = byte_str[:-1].strip()
            try:
                data.append(int(byte_str, 16))
            except ValueError:
                log.warning(f"Invalid byte string {byte_str!r} near offset {parent.offset}")
                return
        yield Submatch(
            f"{self.name}Encoded",
            raw_content,
            relative_offset=0,
            length=len(raw_content),
            parent=parent,
            decoded=bytes(data)
        )


def parse_object(file_stream, object, matcher: Matcher, parent=None):
    log.status('Parsing PDF obj %d %d' % (object.id, object.version))
    objtoken, objid, objversion, endobj = object.objtokens
    pdf_length=endobj.offset.offset - object.content[0].offset.offset + 1 + len(endobj.token)
    if parent is None or isinstance(parent, PDF):
        parent_offset = 0
    else:
        parent_offset = parent.offset
    obj = Submatch(
        name="PDFObject",
        display_name=f"PDFObject{object.id}.{object.version}",
        match_obj=(object.id, object.version),
        relative_offset=objid.offset.offset - parent_offset,
        length=pdf_length + object.content[0].offset.offset - objid.offset.offset,
        parent=parent
    )
    yield obj
    yield Submatch(
        "PDFObjectID",
        object.id,
        relative_offset=0,
        length=len(objid.token),
        parent=obj
    )
    yield Submatch(
        "PDFObjectVersion",
        object.version,
        relative_offset=objversion.offset.offset - objid.offset.offset,
        length=len(objversion.token),
        parent=obj
    )
    log.debug(' Type: %s' % pdfparser.ConditionalCanonicalize(object.GetType(), False))
    log.debug(' Referencing: %s' % ', '.join(map(lambda x: '%s %s %s' % x, object.GetReferences())))
    dataPrecedingStream = object.ContainsStream()
    if dataPrecedingStream:
        log.debug(' Contains stream')
        log.debug(' %s' % pdfparser.FormatOutput(dataPrecedingStream, False))
        oPDFParseDictionary = pdfparser.cPDFParseDictionary(dataPrecedingStream, False)
    else:
        log.debug(' %s' % pdfparser.FormatOutput(object.content, False))
        oPDFParseDictionary = pdfparser.cPDFParseDictionary(object.content, False)
    #log.debug('')
    #pp = BytesIO()
    #oPDFParseDictionary.PrettyPrint('  ', stream=pp)
    #pp.flush()
    #dict_content = pp.read()
    #log.debug(dict_content)
    dict_offset = oPDFParseDictionary.content[0].offset.offset - objid.offset.offset
    dict_length = content_length(oPDFParseDictionary.content)
    if oPDFParseDictionary.parsed is not None:
        yield from _emit_dict(oPDFParseDictionary.parsed, obj, parent.offset)
    #log.debug('')
    #log.debug('')
    content_start = dict_offset + dict_length
    content_len = endobj.offset.offset - content_start - objid.offset.offset
    if content_len > 0:
        content = Submatch(
            "PDFObjectContent",
            (),
            relative_offset=content_start,
            length=content_len,
            parent=obj
        )
        yield content
        stream_len = None
        filters: List[StreamFilter] = []
        if oPDFParseDictionary.parsed is not None:
            if '/Filter' in oPDFParseDictionary.parsed:
                filter_value = oPDFParseDictionary.parsed["/Filter"].strip()
                if filter_value.startswith("[") and filter_value.endswith("]"):
                   filter_value = str(filter_value[1:-1])
                for filter in filter_value.split(" "):
                    if len(filter.strip()) == 0:
                        continue
                    elif filter.strip() not in FILTERS_BY_NAME:
                        log.warn(f"Unimplemented PDF filter: {filter.strip()}")
                    else:
                        new_filter = FILTERS_BY_NAME[filter.strip()]()
                        if filters:
                            filters[-1].next_decoder = new_filter
                        filters.append(new_filter)
            if '/Length' in oPDFParseDictionary.parsed:
                try:
                    stream_len = int(oPDFParseDictionary.parsed['/Length'])
                except ValueError:
                    pass
        old_pos = file_stream.tell()
        try:
            file_stream.seek(content.root_offset)
            raw_content = file_stream.read(content_len)
        finally:
            file_stream.seek(old_pos)
        streamtoken = b'stream'
        if raw_content.startswith(streamtoken):
            raw_content = raw_content[len(streamtoken):]
            if raw_content.startswith(b'\r'):
                streamtoken += b'\r'
                raw_content = raw_content[1:]
            if raw_content.startswith(b'\n'):
                streamtoken += b'\n'
                raw_content = raw_content[1:]
                if raw_content.endswith(b'\n') or raw_content.endswith(b'\r'):
                    endtoken = b'endstream'
                    if raw_content.endswith(b'\r\n'):
                        endtoken += b'\r\n'
                    elif raw_content.endswith(b'\r'):
                        endtoken += b'\r'
                    else:
                        endtoken += b'\n'
                    if raw_content.endswith(endtoken):
                        raw_content = raw_content[:-len(endtoken)]
                        if raw_content.endswith(b'\n') and stream_len is not None and len(raw_content) > stream_len:
                            endtoken = b'\n' + endtoken
                            raw_content = raw_content[:-1]
                        yield Submatch(
                            "StartStream",
                            streamtoken,
                            relative_offset=0,
                            length=len(streamtoken),
                            parent=content
                        )
                        streamcontent = Submatch(
                            "StreamContent",
                            raw_content,
                            relative_offset=len(streamtoken),
                            length=len(raw_content),
                            parent=content
                        )
                        yield streamcontent
                        if filters:
                            yield from filters[0].match(matcher, raw_content, streamcontent)
                        yield Submatch(
                           "EndStream",
                            endtoken,
                            relative_offset=len(streamtoken) + len(raw_content),
                            length=len(endtoken),
                            parent=content
                        )
    log.clear_status()


def parse_pdf(file_stream, matcher: Matcher, parent=None):
    if parent is None or isinstance(parent, PDF):
        parent_offset = 0
    else:
        parent_offset = parent.offset
    with file_stream.tempfile(suffix='.pdf') as pdf_path:
        parser = pdfparser.cPDFParser(pdf_path, True)
        while True:
            object = parser.GetObject()
            if object is None:
                break
            elif object.type == pdfparser.PDF_ELEMENT_COMMENT:
                log.debug(f"PDF comment at {object.offset}, length {len(object.comment)}")
                yield Submatch(
                    name='PDFComment',
                    match_obj=object,
                    relative_offset=object.offset.offset - parent_offset,
                    length=len(object.comment),
                    parent=parent
                )
            elif object.type == pdfparser.PDF_ELEMENT_XREF:
                log.debug('PDF xref')
                yield Submatch(
                    name='PDFXref',
                    match_obj=object,
                    relative_offset=object.content[0].offset.offset - parent_offset,
                    length=content_length(object.content),
                    parent=parent
                )
            elif object.type == pdfparser.PDF_ELEMENT_TRAILER:
                pdfparser.cPDFParseDictionary(object.content[1:], False)
                yield Submatch(
                    name='PDFTrailer',
                    match_obj=object,
                    relative_offset=object.content[0].offset.offset - parent_offset,
                    length=content_length(object.content),
                    parent=parent
                )
            elif object.type == pdfparser.PDF_ELEMENT_STARTXREF:
                yield Submatch(
                    name='PDFStartXRef',
                    match_obj=object.index,
                    relative_offset=object.offset.offset - parent_offset,
                    length=object.length,
                    parent=parent
                )
            elif object.type == pdfparser.PDF_ELEMENT_INDIRECT_OBJECT:
                yield from parse_object(file_stream, object, matcher=matcher, parent=parent)


@submatcher("application/pdf")
class PDF(Match):
    def submatch(self, file_stream):
        yield from parse_pdf(file_stream, matcher=self.matcher, parent=self)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Websocket(KaitaiStruct):
    """The WebSocket protocol establishes a two-way communication channel via TCP.
    Messages are made up of one or more dataframes, and are delineated by
    frames with the `fin` bit set.
    """

    class Opcode(Enum):
        continuation = 0
        text = 1
        binary = 2
        reserved_3 = 3
        reserved_4 = 4
        reserved_5 = 5
        reserved_6 = 6
        reserved_7 = 7
        close = 8
        ping = 9
        pong = 10
        reserved_control_b = 11
        reserved_control_c = 12
        reserved_control_d = 13
        reserved_control_e = 14
        reserved_control_f = 15
    SEQ_FIELDS = ["initial_frame", "trailing_frames"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['initial_frame']['start'] = self._io.pos()
        self.initial_frame = Websocket.InitialFrame(self._io, self, self._root)
        self.initial_frame._read()
        self._debug['initial_frame']['end'] = self._io.pos()
        if self.initial_frame.header.finished != True:
            self._debug['trailing_frames']['start'] = self._io.pos()
            self.trailing_frames = []
            i = 0
            while True:
                if not 'arr' in self._debug['trailing_frames']:
                    self._debug['trailing_frames']['arr'] = []
                self._debug['trailing_frames']['arr'].append({'start': self._io.pos()})
                _t_trailing_frames = Websocket.Dataframe(self._io, self, self._root)
                _t_trailing_frames._read()
                _ = _t_trailing_frames
                self.trailing_frames.append(_)
                self._debug['trailing_frames']['arr'][len(self.trailing_frames) - 1]['end'] = self._io.pos()
                if _.header.finished:
                    break
                i += 1
            self._debug['trailing_frames']['end'] = self._io.pos()


    class FrameHeader(KaitaiStruct):
        SEQ_FIELDS = ["finished", "reserved", "opcode", "is_masked", "len_payload_primary", "len_payload_extended_1", "len_payload_extended_2", "mask_key"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['finished']['start'] = self._io.pos()
            self.finished = self._io.read_bits_int_be(1) != 0
            self._debug['finished']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bits_int_be(3)
            self._debug['reserved']['end'] = self._io.pos()
            self._debug['opcode']['start'] = self._io.pos()
            self.opcode = KaitaiStream.resolve_enum(Websocket.Opcode, self._io.read_bits_int_be(4))
            self._debug['opcode']['end'] = self._io.pos()
            self._debug['is_masked']['start'] = self._io.pos()
            self.is_masked = self._io.read_bits_int_be(1) != 0
            self._debug['is_masked']['end'] = self._io.pos()
            self._debug['len_payload_primary']['start'] = self._io.pos()
            self.len_payload_primary = self._io.read_bits_int_be(7)
            self._debug['len_payload_primary']['end'] = self._io.pos()
            self._io.align_to_byte()
            if self.len_payload_primary == 126:
                self._debug['len_payload_extended_1']['start'] = self._io.pos()
                self.len_payload_extended_1 = self._io.read_u2be()
                self._debug['len_payload_extended_1']['end'] = self._io.pos()

            if self.len_payload_primary == 127:
                self._debug['len_payload_extended_2']['start'] = self._io.pos()
                self.len_payload_extended_2 = self._io.read_u4be()
                self._debug['len_payload_extended_2']['end'] = self._io.pos()

            if self.is_masked:
                self._debug['mask_key']['start'] = self._io.pos()
                self.mask_key = self._io.read_u4be()
                self._debug['mask_key']['end'] = self._io.pos()


        @property
        def len_payload(self):
            if hasattr(self, '_m_len_payload'):
                return self._m_len_payload if hasattr(self, '_m_len_payload') else None

            self._m_len_payload = (self.len_payload_primary if self.len_payload_primary <= 125 else (self.len_payload_extended_1 if self.len_payload_primary == 126 else self.len_payload_extended_2))
            return self._m_len_payload if hasattr(self, '_m_len_payload') else None


    class InitialFrame(KaitaiStruct):
        SEQ_FIELDS = ["header", "payload_bytes", "payload_text"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = Websocket.FrameHeader(self._io, self, self._root)
            self.header._read()
            self._debug['header']['end'] = self._io.pos()
            if self.header.opcode != Websocket.Opcode.text:
                self._debug['payload_bytes']['start'] = self._io.pos()
                self.payload_bytes = self._io.read_bytes(self.header.len_payload)
                self._debug['payload_bytes']['end'] = self._io.pos()

            if self.header.opcode == Websocket.Opcode.text:
                self._debug['payload_text']['start'] = self._io.pos()
                self.payload_text = (self._io.read_bytes(self.header.len_payload)).decode(u"UTF-8")
                self._debug['payload_text']['end'] = self._io.pos()



    class Dataframe(KaitaiStruct):
        SEQ_FIELDS = ["header", "payload_bytes", "payload_text"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = Websocket.FrameHeader(self._io, self, self._root)
            self.header._read()
            self._debug['header']['end'] = self._io.pos()
            if self._root.initial_frame.header.opcode != Websocket.Opcode.text:
                self._debug['payload_bytes']['start'] = self._io.pos()
                self.payload_bytes = self._io.read_bytes(self.header.len_payload)
                self._debug['payload_bytes']['end'] = self._io.pos()

            if self._root.initial_frame.header.opcode == Websocket.Opcode.text:
                self._debug['payload_text']['start'] = self._io.pos()
                self.payload_text = (self._io.read_bytes(self.header.len_payload)).decode(u"UTF-8")
                self._debug['payload_text']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class BroadcomTrx(KaitaiStruct):
    """.trx file format is widely used for distribution of firmware
    updates for Broadcom devices. The most well-known are ASUS routers.
    
    Fundamentally, it includes a footer which acts as a safeguard
    against installing a firmware package on a wrong hardware model or
    version, and a header which list numerous partitions packaged inside
    a single .trx file.
    
    trx files not necessarily contain all these headers.
    
    .. seealso::
       Source - https://github.com/openwrt/openwrt/blob/3f5619f/tools/firmware-utils/src/trx.c
    
    
    .. seealso::
       Source - https://web.archive.org/web/20190127154419/https://openwrt.org/docs/techref/header
    
    
    .. seealso::
       Source - https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/Documentation/devicetree/bindings/mtd/partitions/brcm,trx.txt
    """
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class Revision(KaitaiStruct):
        SEQ_FIELDS = ["major", "minor"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['major']['start'] = self._io.pos()
            self.major = self._io.read_u1()
            self._debug['major']['end'] = self._io.pos()
            self._debug['minor']['start'] = self._io.pos()
            self.minor = self._io.read_u1()
            self._debug['minor']['end'] = self._io.pos()


    class Version(KaitaiStruct):
        SEQ_FIELDS = ["major", "minor", "patch", "tweak"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['major']['start'] = self._io.pos()
            self.major = self._io.read_u1()
            self._debug['major']['end'] = self._io.pos()
            self._debug['minor']['start'] = self._io.pos()
            self.minor = self._io.read_u1()
            self._debug['minor']['end'] = self._io.pos()
            self._debug['patch']['start'] = self._io.pos()
            self.patch = self._io.read_u1()
            self._debug['patch']['end'] = self._io.pos()
            self._debug['tweak']['start'] = self._io.pos()
            self.tweak = self._io.read_u1()
            self._debug['tweak']['end'] = self._io.pos()


    class Tail(KaitaiStruct):
        """A safeguard against installation of firmware to an incompatible device."""
        SEQ_FIELDS = ["version", "product_id", "comp_hw", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = BroadcomTrx.Version(self._io, self, self._root)
            self.version._read()
            self._debug['version']['end'] = self._io.pos()
            self._debug['product_id']['start'] = self._io.pos()
            self.product_id = (KaitaiStream.bytes_terminate(self._io.read_bytes(12), 0, False)).decode(u"utf-8")
            self._debug['product_id']['end'] = self._io.pos()
            self._debug['comp_hw']['start'] = self._io.pos()
            self.comp_hw = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['comp_hw']:
                    self._debug['comp_hw']['arr'] = []
                self._debug['comp_hw']['arr'].append({'start': self._io.pos()})
                _t_comp_hw = BroadcomTrx.Tail.HwCompInfo(self._io, self, self._root)
                _t_comp_hw._read()
                self.comp_hw[i] = _t_comp_hw
                self._debug['comp_hw']['arr'][i]['end'] = self._io.pos()

            self._debug['comp_hw']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(32)
            self._debug['reserved']['end'] = self._io.pos()

        class HwCompInfo(KaitaiStruct):
            SEQ_FIELDS = ["min", "max"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['min']['start'] = self._io.pos()
                self.min = BroadcomTrx.Revision(self._io, self, self._root)
                self.min._read()
                self._debug['min']['end'] = self._io.pos()
                self._debug['max']['start'] = self._io.pos()
                self.max = BroadcomTrx.Revision(self._io, self, self._root)
                self.max._read()
                self._debug['max']['end'] = self._io.pos()



    class Header(KaitaiStruct):
        SEQ_FIELDS = ["magic", "len", "crc32", "version", "flags", "partitions"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x48\x44\x52\x30":
                raise kaitaistruct.ValidationNotEqualError(b"\x48\x44\x52\x30", self.magic, self._io, u"/types/header/seq/0")
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u4le()
            self._debug['len']['end'] = self._io.pos()
            self._debug['crc32']['start'] = self._io.pos()
            self.crc32 = self._io.read_u4le()
            self._debug['crc32']['end'] = self._io.pos()
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u2le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = BroadcomTrx.Header.Flags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['partitions']['start'] = self._io.pos()
            self.partitions = []
            i = 0
            while True:
                if not 'arr' in self._debug['partitions']:
                    self._debug['partitions']['arr'] = []
                self._debug['partitions']['arr'].append({'start': self._io.pos()})
                _t_partitions = BroadcomTrx.Header.Partition(i, self._io, self, self._root)
                _t_partitions._read()
                _ = _t_partitions
                self.partitions.append(_)
                self._debug['partitions']['arr'][len(self.partitions) - 1]['end'] = self._io.pos()
                if  ((i >= 4) or (not (_.is_present))) :
                    break
                i += 1
            self._debug['partitions']['end'] = self._io.pos()

        class Partition(KaitaiStruct):
            SEQ_FIELDS = ["ofs_body"]
            def __init__(self, idx, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self.idx = idx
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['ofs_body']['start'] = self._io.pos()
                self.ofs_body = self._io.read_u4le()
                self._debug['ofs_body']['end'] = self._io.pos()

            @property
            def is_present(self):
                if hasattr(self, '_m_is_present'):
                    return self._m_is_present if hasattr(self, '_m_is_present') else None

                self._m_is_present = self.ofs_body != 0
                return self._m_is_present if hasattr(self, '_m_is_present') else None

            @property
            def is_last(self):
                if hasattr(self, '_m_is_last'):
                    return self._m_is_last if hasattr(self, '_m_is_last') else None

                if self.is_present:
                    self._m_is_last =  ((self.idx == (len(self._parent.partitions) - 1)) or (not (self._parent.partitions[(self.idx + 1)].is_present))) 

                return self._m_is_last if hasattr(self, '_m_is_last') else None

            @property
            def len_body(self):
                if hasattr(self, '_m_len_body'):
                    return self._m_len_body if hasattr(self, '_m_len_body') else None

                if self.is_present:
                    self._m_len_body = ((self._root._io.size() - self.ofs_body) if self.is_last else self._parent.partitions[(self.idx + 1)].ofs_body)

                return self._m_len_body if hasattr(self, '_m_len_body') else None

            @property
            def body(self):
                if hasattr(self, '_m_body'):
                    return self._m_body if hasattr(self, '_m_body') else None

                if self.is_present:
                    io = self._root._io
                    _pos = io.pos()
                    io.seek(self.ofs_body)
                    self._debug['_m_body']['start'] = io.pos()
                    self._m_body = io.read_bytes(self.len_body)
                    self._debug['_m_body']['end'] = io.pos()
                    io.seek(_pos)

                return self._m_body if hasattr(self, '_m_body') else None


        class Flags(KaitaiStruct):
            SEQ_FIELDS = ["flags"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['flags']['start'] = self._io.pos()
                self.flags = [None] * (16)
                for i in range(16):
                    if not 'arr' in self._debug['flags']:
                        self._debug['flags']['arr'] = []
                    self._debug['flags']['arr'].append({'start': self._io.pos()})
                    self.flags[i] = self._io.read_bits_int_le(1) != 0
                    self._debug['flags']['arr'][i]['end'] = self._io.pos()

                self._debug['flags']['end'] = self._io.pos()



    @property
    def header(self):
        if hasattr(self, '_m_header'):
            return self._m_header if hasattr(self, '_m_header') else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._debug['_m_header']['start'] = self._io.pos()
        self._m_header = BroadcomTrx.Header(self._io, self, self._root)
        self._m_header._read()
        self._debug['_m_header']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_header if hasattr(self, '_m_header') else None

    @property
    def tail(self):
        if hasattr(self, '_m_tail'):
            return self._m_tail if hasattr(self, '_m_tail') else None

        _pos = self._io.pos()
        self._io.seek((self._io.size() - 64))
        self._debug['_m_tail']['start'] = self._io.pos()
        self._m_tail = BroadcomTrx.Tail(self._io, self, self._root)
        self._m_tail._read()
        self._debug['_m_tail']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_tail if hasattr(self, '_m_tail') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class MagicavoxelVox(KaitaiStruct):
    """
    .. seealso::
       MagicaVoxel Homepage - https://ephtracy.github.io/
    
    
    .. seealso::
       Format Description - https://github.com/ephtracy/voxel-model/blob/master/MagicaVoxel-file-format-vox.txt
    """

    class ChunkType(Enum):
        main = 1296124238
        matt = 1296127060
        pack = 1346454347
        rgba = 1380401729
        size = 1397316165
        xyzi = 1482250825

    class MaterialType(Enum):
        diffuse = 0
        metal = 1
        glass = 2
        emissive = 3

    class PropertyBitsType(Enum):
        plastic = 1
        roughness = 2
        specular = 4
        ior = 8
        attenuation = 16
        power = 32
        glow = 64
        is_total_power = 128
    SEQ_FIELDS = ["magic", "version", "main"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(4)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x56\x4F\x58\x20":
            raise kaitaistruct.ValidationNotEqualError(b"\x56\x4F\x58\x20", self.magic, self._io, u"/seq/0")
        self._debug['version']['start'] = self._io.pos()
        self.version = self._io.read_u4le()
        self._debug['version']['end'] = self._io.pos()
        self._debug['main']['start'] = self._io.pos()
        self.main = MagicavoxelVox.Chunk(self._io, self, self._root)
        self.main._read()
        self._debug['main']['end'] = self._io.pos()

    class Chunk(KaitaiStruct):
        SEQ_FIELDS = ["chunk_id", "num_bytes_of_chunk_content", "num_bytes_of_children_chunks", "chunk_content", "children_chunks"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['chunk_id']['start'] = self._io.pos()
            self.chunk_id = KaitaiStream.resolve_enum(MagicavoxelVox.ChunkType, self._io.read_u4be())
            self._debug['chunk_id']['end'] = self._io.pos()
            self._debug['num_bytes_of_chunk_content']['start'] = self._io.pos()
            self.num_bytes_of_chunk_content = self._io.read_u4le()
            self._debug['num_bytes_of_chunk_content']['end'] = self._io.pos()
            self._debug['num_bytes_of_children_chunks']['start'] = self._io.pos()
            self.num_bytes_of_children_chunks = self._io.read_u4le()
            self._debug['num_bytes_of_children_chunks']['end'] = self._io.pos()
            if self.num_bytes_of_chunk_content != 0:
                self._debug['chunk_content']['start'] = self._io.pos()
                _on = self.chunk_id
                if _on == MagicavoxelVox.ChunkType.size:
                    self._raw_chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                    _io__raw_chunk_content = KaitaiStream(BytesIO(self._raw_chunk_content))
                    self.chunk_content = MagicavoxelVox.Size(_io__raw_chunk_content, self, self._root)
                    self.chunk_content._read()
                elif _on == MagicavoxelVox.ChunkType.matt:
                    self._raw_chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                    _io__raw_chunk_content = KaitaiStream(BytesIO(self._raw_chunk_content))
                    self.chunk_content = MagicavoxelVox.Matt(_io__raw_chunk_content, self, self._root)
                    self.chunk_content._read()
                elif _on == MagicavoxelVox.ChunkType.rgba:
                    self._raw_chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                    _io__raw_chunk_content = KaitaiStream(BytesIO(self._raw_chunk_content))
                    self.chunk_content = MagicavoxelVox.Rgba(_io__raw_chunk_content, self, self._root)
                    self.chunk_content._read()
                elif _on == MagicavoxelVox.ChunkType.xyzi:
                    self._raw_chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                    _io__raw_chunk_content = KaitaiStream(BytesIO(self._raw_chunk_content))
                    self.chunk_content = MagicavoxelVox.Xyzi(_io__raw_chunk_content, self, self._root)
                    self.chunk_content._read()
                elif _on == MagicavoxelVox.ChunkType.pack:
                    self._raw_chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                    _io__raw_chunk_content = KaitaiStream(BytesIO(self._raw_chunk_content))
                    self.chunk_content = MagicavoxelVox.Pack(_io__raw_chunk_content, self, self._root)
                    self.chunk_content._read()
                else:
                    self.chunk_content = self._io.read_bytes(self.num_bytes_of_chunk_content)
                self._debug['chunk_content']['end'] = self._io.pos()

            if self.num_bytes_of_children_chunks != 0:
                self._debug['children_chunks']['start'] = self._io.pos()
                self.children_chunks = []
                i = 0
                while not self._io.is_eof():
                    if not 'arr' in self._debug['children_chunks']:
                        self._debug['children_chunks']['arr'] = []
                    self._debug['children_chunks']['arr'].append({'start': self._io.pos()})
                    _t_children_chunks = MagicavoxelVox.Chunk(self._io, self, self._root)
                    _t_children_chunks._read()
                    self.children_chunks.append(_t_children_chunks)
                    self._debug['children_chunks']['arr'][len(self.children_chunks) - 1]['end'] = self._io.pos()
                    i += 1

                self._debug['children_chunks']['end'] = self._io.pos()



    class Size(KaitaiStruct):
        SEQ_FIELDS = ["size_x", "size_y", "size_z"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size_x']['start'] = self._io.pos()
            self.size_x = self._io.read_u4le()
            self._debug['size_x']['end'] = self._io.pos()
            self._debug['size_y']['start'] = self._io.pos()
            self.size_y = self._io.read_u4le()
            self._debug['size_y']['end'] = self._io.pos()
            self._debug['size_z']['start'] = self._io.pos()
            self.size_z = self._io.read_u4le()
            self._debug['size_z']['end'] = self._io.pos()


    class Rgba(KaitaiStruct):
        SEQ_FIELDS = ["colors"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['colors']['start'] = self._io.pos()
            self.colors = [None] * (256)
            for i in range(256):
                if not 'arr' in self._debug['colors']:
                    self._debug['colors']['arr'] = []
                self._debug['colors']['arr'].append({'start': self._io.pos()})
                _t_colors = MagicavoxelVox.Color(self._io, self, self._root)
                _t_colors._read()
                self.colors[i] = _t_colors
                self._debug['colors']['arr'][i]['end'] = self._io.pos()

            self._debug['colors']['end'] = self._io.pos()


    class Pack(KaitaiStruct):
        SEQ_FIELDS = ["num_models"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_models']['start'] = self._io.pos()
            self.num_models = self._io.read_u4le()
            self._debug['num_models']['end'] = self._io.pos()


    class Matt(KaitaiStruct):
        SEQ_FIELDS = ["id", "material_type", "material_weight", "property_bits", "plastic", "roughness", "specular", "ior", "attenuation", "power", "glow", "is_total_power"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['id']['start'] = self._io.pos()
            self.id = self._io.read_u4le()
            self._debug['id']['end'] = self._io.pos()
            self._debug['material_type']['start'] = self._io.pos()
            self.material_type = KaitaiStream.resolve_enum(MagicavoxelVox.MaterialType, self._io.read_u4le())
            self._debug['material_type']['end'] = self._io.pos()
            self._debug['material_weight']['start'] = self._io.pos()
            self.material_weight = self._io.read_f4le()
            self._debug['material_weight']['end'] = self._io.pos()
            self._debug['property_bits']['start'] = self._io.pos()
            self.property_bits = self._io.read_u4le()
            self._debug['property_bits']['end'] = self._io.pos()
            if self.has_plastic:
                self._debug['plastic']['start'] = self._io.pos()
                self.plastic = self._io.read_f4le()
                self._debug['plastic']['end'] = self._io.pos()

            if self.has_roughness:
                self._debug['roughness']['start'] = self._io.pos()
                self.roughness = self._io.read_f4le()
                self._debug['roughness']['end'] = self._io.pos()

            if self.has_specular:
                self._debug['specular']['start'] = self._io.pos()
                self.specular = self._io.read_f4le()
                self._debug['specular']['end'] = self._io.pos()

            if self.has_ior:
                self._debug['ior']['start'] = self._io.pos()
                self.ior = self._io.read_f4le()
                self._debug['ior']['end'] = self._io.pos()

            if self.has_attenuation:
                self._debug['attenuation']['start'] = self._io.pos()
                self.attenuation = self._io.read_f4le()
                self._debug['attenuation']['end'] = self._io.pos()

            if self.has_power:
                self._debug['power']['start'] = self._io.pos()
                self.power = self._io.read_f4le()
                self._debug['power']['end'] = self._io.pos()

            if self.has_glow:
                self._debug['glow']['start'] = self._io.pos()
                self.glow = self._io.read_f4le()
                self._debug['glow']['end'] = self._io.pos()

            if self.has_is_total_power:
                self._debug['is_total_power']['start'] = self._io.pos()
                self.is_total_power = self._io.read_f4le()
                self._debug['is_total_power']['end'] = self._io.pos()


        @property
        def has_is_total_power(self):
            if hasattr(self, '_m_has_is_total_power'):
                return self._m_has_is_total_power if hasattr(self, '_m_has_is_total_power') else None

            self._m_has_is_total_power = (self.property_bits & 128) != 0
            return self._m_has_is_total_power if hasattr(self, '_m_has_is_total_power') else None

        @property
        def has_plastic(self):
            if hasattr(self, '_m_has_plastic'):
                return self._m_has_plastic if hasattr(self, '_m_has_plastic') else None

            self._m_has_plastic = (self.property_bits & 1) != 0
            return self._m_has_plastic if hasattr(self, '_m_has_plastic') else None

        @property
        def has_attenuation(self):
            if hasattr(self, '_m_has_attenuation'):
                return self._m_has_attenuation if hasattr(self, '_m_has_attenuation') else None

            self._m_has_attenuation = (self.property_bits & 16) != 0
            return self._m_has_attenuation if hasattr(self, '_m_has_attenuation') else None

        @property
        def has_power(self):
            if hasattr(self, '_m_has_power'):
                return self._m_has_power if hasattr(self, '_m_has_power') else None

            self._m_has_power = (self.property_bits & 32) != 0
            return self._m_has_power if hasattr(self, '_m_has_power') else None

        @property
        def has_roughness(self):
            if hasattr(self, '_m_has_roughness'):
                return self._m_has_roughness if hasattr(self, '_m_has_roughness') else None

            self._m_has_roughness = (self.property_bits & 2) != 0
            return self._m_has_roughness if hasattr(self, '_m_has_roughness') else None

        @property
        def has_specular(self):
            if hasattr(self, '_m_has_specular'):
                return self._m_has_specular if hasattr(self, '_m_has_specular') else None

            self._m_has_specular = (self.property_bits & 4) != 0
            return self._m_has_specular if hasattr(self, '_m_has_specular') else None

        @property
        def has_ior(self):
            if hasattr(self, '_m_has_ior'):
                return self._m_has_ior if hasattr(self, '_m_has_ior') else None

            self._m_has_ior = (self.property_bits & 8) != 0
            return self._m_has_ior if hasattr(self, '_m_has_ior') else None

        @property
        def has_glow(self):
            if hasattr(self, '_m_has_glow'):
                return self._m_has_glow if hasattr(self, '_m_has_glow') else None

            self._m_has_glow = (self.property_bits & 64) != 0
            return self._m_has_glow if hasattr(self, '_m_has_glow') else None


    class Xyzi(KaitaiStruct):
        SEQ_FIELDS = ["num_voxels", "voxels"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_voxels']['start'] = self._io.pos()
            self.num_voxels = self._io.read_u4le()
            self._debug['num_voxels']['end'] = self._io.pos()
            self._debug['voxels']['start'] = self._io.pos()
            self.voxels = [None] * (self.num_voxels)
            for i in range(self.num_voxels):
                if not 'arr' in self._debug['voxels']:
                    self._debug['voxels']['arr'] = []
                self._debug['voxels']['arr'].append({'start': self._io.pos()})
                _t_voxels = MagicavoxelVox.Voxel(self._io, self, self._root)
                _t_voxels._read()
                self.voxels[i] = _t_voxels
                self._debug['voxels']['arr'][i]['end'] = self._io.pos()

            self._debug['voxels']['end'] = self._io.pos()


    class Color(KaitaiStruct):
        SEQ_FIELDS = ["r", "g", "b", "a"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['r']['start'] = self._io.pos()
            self.r = self._io.read_u1()
            self._debug['r']['end'] = self._io.pos()
            self._debug['g']['start'] = self._io.pos()
            self.g = self._io.read_u1()
            self._debug['g']['end'] = self._io.pos()
            self._debug['b']['start'] = self._io.pos()
            self.b = self._io.read_u1()
            self._debug['b']['end'] = self._io.pos()
            self._debug['a']['start'] = self._io.pos()
            self.a = self._io.read_u1()
            self._debug['a']['end'] = self._io.pos()


    class Voxel(KaitaiStruct):
        SEQ_FIELDS = ["x", "y", "z", "color_index"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['x']['start'] = self._io.pos()
            self.x = self._io.read_u1()
            self._debug['x']['end'] = self._io.pos()
            self._debug['y']['start'] = self._io.pos()
            self.y = self._io.read_u1()
            self._debug['y']['end'] = self._io.pos()
            self._debug['z']['start'] = self._io.pos()
            self.z = self._io.read_u1()
            self._debug['z']['end'] = self._io.pos()
            self._debug['color_index']['start'] = self._io.pos()
            self.color_index = self._io.read_u1()
            self._debug['color_index']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AllegroDat(KaitaiStruct):
    """Allegro library for C (mostly used for game and multimedia apps
    programming) used its own container file format.
    
    In general, it allows storage of arbitrary binary data blocks
    bundled together with some simple key-value style metadata
    ("properties") for every block. Allegro also pre-defines some simple
    formats for bitmaps, fonts, MIDI music, sound samples and
    palettes. Allegro library v4.0+ also support LZSS compression.
    
    This spec applies to Allegro data files for library versions 2.2 up
    to 4.4.
    
    .. seealso::
       Source - https://liballeg.org/stabledocs/en/datafile.html
    """

    class PackEnum(Enum):
        unpacked = 1936484398
    SEQ_FIELDS = ["pack_magic", "dat_magic", "num_objects", "objects"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['pack_magic']['start'] = self._io.pos()
        self.pack_magic = KaitaiStream.resolve_enum(AllegroDat.PackEnum, self._io.read_u4be())
        self._debug['pack_magic']['end'] = self._io.pos()
        self._debug['dat_magic']['start'] = self._io.pos()
        self.dat_magic = self._io.read_bytes(4)
        self._debug['dat_magic']['end'] = self._io.pos()
        if not self.dat_magic == b"\x41\x4C\x4C\x2E":
            raise kaitaistruct.ValidationNotEqualError(b"\x41\x4C\x4C\x2E", self.dat_magic, self._io, u"/seq/1")
        self._debug['num_objects']['start'] = self._io.pos()
        self.num_objects = self._io.read_u4be()
        self._debug['num_objects']['end'] = self._io.pos()
        self._debug['objects']['start'] = self._io.pos()
        self.objects = [None] * (self.num_objects)
        for i in range(self.num_objects):
            if not 'arr' in self._debug['objects']:
                self._debug['objects']['arr'] = []
            self._debug['objects']['arr'].append({'start': self._io.pos()})
            _t_objects = AllegroDat.DatObject(self._io, self, self._root)
            _t_objects._read()
            self.objects[i] = _t_objects
            self._debug['objects']['arr'][i]['end'] = self._io.pos()

        self._debug['objects']['end'] = self._io.pos()

    class DatFont16(KaitaiStruct):
        """Simple monochrome monospaced font, 95 characters, 8x16 px
        characters.
        """
        SEQ_FIELDS = ["chars"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['chars']['start'] = self._io.pos()
            self.chars = [None] * (95)
            for i in range(95):
                if not 'arr' in self._debug['chars']:
                    self._debug['chars']['arr'] = []
                self._debug['chars']['arr'].append({'start': self._io.pos()})
                self.chars[i] = self._io.read_bytes(16)
                self._debug['chars']['arr'][i]['end'] = self._io.pos()

            self._debug['chars']['end'] = self._io.pos()


    class DatBitmap(KaitaiStruct):
        SEQ_FIELDS = ["bits_per_pixel", "width", "height", "image"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['bits_per_pixel']['start'] = self._io.pos()
            self.bits_per_pixel = self._io.read_s2be()
            self._debug['bits_per_pixel']['end'] = self._io.pos()
            self._debug['width']['start'] = self._io.pos()
            self.width = self._io.read_u2be()
            self._debug['width']['end'] = self._io.pos()
            self._debug['height']['start'] = self._io.pos()
            self.height = self._io.read_u2be()
            self._debug['height']['end'] = self._io.pos()
            self._debug['image']['start'] = self._io.pos()
            self.image = self._io.read_bytes_full()
            self._debug['image']['end'] = self._io.pos()


    class DatFont(KaitaiStruct):
        SEQ_FIELDS = ["font_size", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['font_size']['start'] = self._io.pos()
            self.font_size = self._io.read_s2be()
            self._debug['font_size']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.font_size
            if _on == 8:
                self.body = AllegroDat.DatFont8(self._io, self, self._root)
                self.body._read()
            elif _on == 16:
                self.body = AllegroDat.DatFont16(self._io, self, self._root)
                self.body._read()
            elif _on == 0:
                self.body = AllegroDat.DatFont39(self._io, self, self._root)
                self.body._read()
            self._debug['body']['end'] = self._io.pos()


    class DatFont8(KaitaiStruct):
        """Simple monochrome monospaced font, 95 characters, 8x8 px
        characters.
        """
        SEQ_FIELDS = ["chars"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['chars']['start'] = self._io.pos()
            self.chars = [None] * (95)
            for i in range(95):
                if not 'arr' in self._debug['chars']:
                    self._debug['chars']['arr'] = []
                self._debug['chars']['arr'].append({'start': self._io.pos()})
                self.chars[i] = self._io.read_bytes(8)
                self._debug['chars']['arr'][i]['end'] = self._io.pos()

            self._debug['chars']['end'] = self._io.pos()


    class DatObject(KaitaiStruct):
        SEQ_FIELDS = ["properties", "len_compressed", "len_uncompressed", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['properties']['start'] = self._io.pos()
            self.properties = []
            i = 0
            while True:
                if not 'arr' in self._debug['properties']:
                    self._debug['properties']['arr'] = []
                self._debug['properties']['arr'].append({'start': self._io.pos()})
                _t_properties = AllegroDat.Property(self._io, self, self._root)
                _t_properties._read()
                _ = _t_properties
                self.properties.append(_)
                self._debug['properties']['arr'][len(self.properties) - 1]['end'] = self._io.pos()
                if not (_.is_valid):
                    break
                i += 1
            self._debug['properties']['end'] = self._io.pos()
            self._debug['len_compressed']['start'] = self._io.pos()
            self.len_compressed = self._io.read_s4be()
            self._debug['len_compressed']['end'] = self._io.pos()
            self._debug['len_uncompressed']['start'] = self._io.pos()
            self.len_uncompressed = self._io.read_s4be()
            self._debug['len_uncompressed']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.type
            if _on == u"BMP ":
                self._raw_body = self._io.read_bytes(self.len_compressed)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = AllegroDat.DatBitmap(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == u"RLE ":
                self._raw_body = self._io.read_bytes(self.len_compressed)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = AllegroDat.DatRleSprite(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == u"FONT":
                self._raw_body = self._io.read_bytes(self.len_compressed)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = AllegroDat.DatFont(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.len_compressed)
            self._debug['body']['end'] = self._io.pos()

        @property
        def type(self):
            if hasattr(self, '_m_type'):
                return self._m_type if hasattr(self, '_m_type') else None

            self._m_type = self.properties[-1].magic
            return self._m_type if hasattr(self, '_m_type') else None


    class DatFont39(KaitaiStruct):
        """New bitmap font format introduced since Allegro 3.9: allows
        flexible designation of character ranges, 8-bit colored
        characters, etc.
        """
        SEQ_FIELDS = ["num_ranges", "ranges"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_ranges']['start'] = self._io.pos()
            self.num_ranges = self._io.read_s2be()
            self._debug['num_ranges']['end'] = self._io.pos()
            self._debug['ranges']['start'] = self._io.pos()
            self.ranges = [None] * (self.num_ranges)
            for i in range(self.num_ranges):
                if not 'arr' in self._debug['ranges']:
                    self._debug['ranges']['arr'] = []
                self._debug['ranges']['arr'].append({'start': self._io.pos()})
                _t_ranges = AllegroDat.DatFont39.Range(self._io, self, self._root)
                _t_ranges._read()
                self.ranges[i] = _t_ranges
                self._debug['ranges']['arr'][i]['end'] = self._io.pos()

            self._debug['ranges']['end'] = self._io.pos()

        class Range(KaitaiStruct):
            SEQ_FIELDS = ["mono", "start_char", "end_char", "chars"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['mono']['start'] = self._io.pos()
                self.mono = self._io.read_u1()
                self._debug['mono']['end'] = self._io.pos()
                self._debug['start_char']['start'] = self._io.pos()
                self.start_char = self._io.read_u4be()
                self._debug['start_char']['end'] = self._io.pos()
                self._debug['end_char']['start'] = self._io.pos()
                self.end_char = self._io.read_u4be()
                self._debug['end_char']['end'] = self._io.pos()
                self._debug['chars']['start'] = self._io.pos()
                self.chars = [None] * (((self.end_char - self.start_char) + 1))
                for i in range(((self.end_char - self.start_char) + 1)):
                    if not 'arr' in self._debug['chars']:
                        self._debug['chars']['arr'] = []
                    self._debug['chars']['arr'].append({'start': self._io.pos()})
                    _t_chars = AllegroDat.DatFont39.FontChar(self._io, self, self._root)
                    _t_chars._read()
                    self.chars[i] = _t_chars
                    self._debug['chars']['arr'][i]['end'] = self._io.pos()

                self._debug['chars']['end'] = self._io.pos()


        class FontChar(KaitaiStruct):
            SEQ_FIELDS = ["width", "height", "body"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['width']['start'] = self._io.pos()
                self.width = self._io.read_u2be()
                self._debug['width']['end'] = self._io.pos()
                self._debug['height']['start'] = self._io.pos()
                self.height = self._io.read_u2be()
                self._debug['height']['end'] = self._io.pos()
                self._debug['body']['start'] = self._io.pos()
                self.body = self._io.read_bytes((self.width * self.height))
                self._debug['body']['end'] = self._io.pos()



    class Property(KaitaiStruct):
        SEQ_FIELDS = ["magic", "type", "len_body", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = (self._io.read_bytes(4)).decode(u"UTF-8")
            self._debug['magic']['end'] = self._io.pos()
            if self.is_valid:
                self._debug['type']['start'] = self._io.pos()
                self.type = (self._io.read_bytes(4)).decode(u"UTF-8")
                self._debug['type']['end'] = self._io.pos()

            if self.is_valid:
                self._debug['len_body']['start'] = self._io.pos()
                self.len_body = self._io.read_u4be()
                self._debug['len_body']['end'] = self._io.pos()

            if self.is_valid:
                self._debug['body']['start'] = self._io.pos()
                self.body = (self._io.read_bytes(self.len_body)).decode(u"UTF-8")
                self._debug['body']['end'] = self._io.pos()


        @property
        def is_valid(self):
            if hasattr(self, '_m_is_valid'):
                return self._m_is_valid if hasattr(self, '_m_is_valid') else None

            self._m_is_valid = self.magic == u"prop"
            return self._m_is_valid if hasattr(self, '_m_is_valid') else None


    class DatRleSprite(KaitaiStruct):
        SEQ_FIELDS = ["bits_per_pixel", "width", "height", "len_image", "image"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['bits_per_pixel']['start'] = self._io.pos()
            self.bits_per_pixel = self._io.read_s2be()
            self._debug['bits_per_pixel']['end'] = self._io.pos()
            self._debug['width']['start'] = self._io.pos()
            self.width = self._io.read_u2be()
            self._debug['width']['end'] = self._io.pos()
            self._debug['height']['start'] = self._io.pos()
            self.height = self._io.read_u2be()
            self._debug['height']['end'] = self._io.pos()
            self._debug['len_image']['start'] = self._io.pos()
            self.len_image = self._io.read_u4be()
            self._debug['len_image']['end'] = self._io.pos()
            self._debug['image']['start'] = self._io.pos()
            self.image = self._io.read_bytes_full()
            self._debug['image']['end'] = self._io.pos()



